#!/usr/bin/perl -w

# Copyright 2014, 2015, 2016, 2017, 2018, 2019, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.010;
use strict;
use Carp 'croak';
use File::Spec;
use List::Util 'max','min','sum','pairmap';
use Math::BaseCnv 'cnv';
use Math::Complex 'pi';
use Math::Libm 'hypot';
use POSIX 'floor';

use Math::PlanePath::R5DragonCurve;
use Math::PlanePath::DragonCurve;

use Math::NumSeq::PlanePathDelta;
BEGIN { *dxdy_to_dir4 = \&Math::NumSeq::PlanePathDelta::_dxdy_to_dir4; }

use FindBin;
use lib::abs "$FindBin::Bin/../tools";
use MyDragon ':all';
use MyPlanar;
our @ipc;
$|=1;

# uncomment this to run the ### lines
# use Smart::Comments;

# Knuth counts of folding rules following Dekking
# cf Jorg Arndt square grid distinct shapes A265685


{
  # A117633
  # first step street or avenue so 2 paths
  # n=4 less than 2^n since cannot go around a block to self-contact
  # dir:=[[(-1)**y, 0], [0, (-1)**x]]
  # even y rows go East, odd y rows go West
  # even x columns go North, odd x columns go South

  my $verbose = 0;
  my @walks = ([ [0,0] ]);
  foreach my $n (2 .. 12) {
    my @new_walks;
    my %seen;
    foreach my $walk (@walks) {
      foreach my $dir4 (0 .. 3) {
        ### $dir4
        my $end = $walk->[-1];
        foreach my $dp ([(-1)**$end->[1], 0],
                        [0, (-1)**$end->[0]]) {
          my $p = [ $end->[0] + $dp->[0],
                    $end->[1] + $dp->[1] ];
          next if points_contains_point($walk,$p);
          my $new_walk = [ @$walk, $p ];
          # $new_walk->[1]->[0] == 1 or next;
          # $new_walk->[1]->[1] == 0 or next;

          my $key = points_str($new_walk);
          ### $key
          if (! $seen{$key}) {
            $seen{$key} = 1;
            push @new_walks, $new_walk;
          }
        }
      }
    }
    @walks = @new_walks;
    my $count = scalar(@walks);
    if ($verbose) {
      print "n=$n count $count\n";
    } else {
      print "$count,";
    }
  }
  exit 0;
}

{
  #     1,2,4,7,13,24,44,77,139,250,450,788,1403,2498,4447,7782,13769,24363,43106,
  # n = 1 2 3 4
  # A006744 self-avoiding walks n steps
  #
  # A117633
  # first step street or avenue so 2 paths
  # n=4 less than 2^n since cannot go around a block to self-contact
  # dir:=[[(-1)**y, 0], [0, (-1)**x]]
  # go x or y
  #
  # A046661 n-step self-avoiding walks first step specified
  # start East then any
  #
  # A046171 n-step self-avoiding walks, if any verticals then go up before any down
  # up to mirror image

  my $verbose = 0;
  my @walks = ([ [0,0] ]);
  foreach my $n (2 .. 8) {
    my @new_walks;
    my %seen;
    foreach my $walk (@walks) {
      foreach my $dir4 (0 .. 3) {
        ### $dir4
        my $end = $walk->[-1];
        my $p = [$end->[0] + $dir4_to_dx[$dir4],
                 $end->[1] + $dir4_to_dy[$dir4]];
        next if points_contains_point($walk,$p);
        my $new_walk = [ @$walk, $p ];
        $new_walk->[1]->[0] == 1 or next;
        $new_walk->[1]->[1] == 0 or next;
        
        my $key = points_str($new_walk);
        ### $key
        if (! $seen{$key}) {
          my $conj_walk = MyPlanar::points_xyscale($new_walk, 1, -1);
          my $conj_key = points_str($conj_walk);
          if (! $seen{$conj_key}) {
            $seen{$key} = 1;
            $seen{$conj_key} = 1;
            push @new_walks, $new_walk;
          }
        }
      }
    }
    @walks = @new_walks;
    my $count = scalar(@walks);
    if ($verbose) {
      print "n=$n count $count\n";
    } else {
      print "$count,";
    }
  }
  exit 0;

  sub points_contains_point {
    my ($points, $point) = @_;
    foreach my $i (0 .. $#$points) {
      if ($points->[$i]->[0] == $point->[0]
          && $points->[$i]->[1] == $point->[1]) {
        return 1;
      }
    }
    return 0;
  }
}

{
  # num paths of length s starting East and first turn L which end at norm=s
  # not in OEIS: 1,0,2,3,0,0,16,15,100,0,0,315,0,0,1960,3528,7056,0,50400,
  #              s=2 3 4
  # not in OEIS: 1,2,3,16,15,100,315,1960,3528,7056,50400,
  #
  #     *        *
  #     |        |
  # O---L    O---L

  # my(v=vector(50,i,1)); for(x=0,50, for(y=0,50, my(n=x^2+y^2); if(n>0 && n<=#v, v[n]=0))); v
  # norms pred A229062
  # 1, 1, 0, 1, 1, 0, 0, 1, 1,  1, 0, 0, 1, 0,0,    1,   1,  1,  0, 1,   0,0,0,0,  1,     1,     0, 0, 1, 0, 0, 1, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 0, 0, 1, 0, 0, 0, 1, 1
  # 1, 1, 0, 2, 3, 0, 0,16,15,100, 0, 0,315,0,0,  1960,3528,7056,0,50400,0,0,0,0,967395,2453880,
  # 1  2  3  4  5  6  7  8  9  10  11 12 13        16   17   18      20
  #
  # norms values A001481  more 4k+1 divisors than 4k+3 divisors
  # 1, 2, 4, 5, 8, 9, 10, 13, 16, 17, 18, 20, 25, 26, 29, 32, 34, 36, 37, 40, 41, 45, 49, 50
  # non-norms values A022544
  # my(m=Map()); for(x=0,50, for(y=0,50, my(n=x^2+y^2); if(n<=50, mapput(m,n,1)))); \
  # my(v=Mat(m)[,1]~); for(i=1,#v,print(factor(v[i])))
  #
  # endpoint not norm n
  #          n = 2 3 4 5  6  7  8   9
  # not in OEIS: 0,2,2,5,16,32,48,113,156,512,1024,1733,4096,8192,14424,29240,58480,131072,211744,
  #
  # base figure overlap occurs
  #          n = 2 3 4 5  6  7  8   9
  # not in OEIS: 0,0,0,1,3,8,20,51,117,262,574,1260,2693,5694,11937,24986,51767,
  # not in OEIS:       1,3,8,20,51,117,262,574,1260,2693,5694,11937,
  #
  # base figure no overlap occurs
  # n = 2 3 4 5  6  7  8  9
  #     1,2,4,7,13,24,44,77,139,250,450,788,1403,2498,4447,7782,13769,24363,43106,
  # n = 1 2 3 4
  # A006744 self-avoiding walks n steps
  # always 90 degrees take midpoints to give any left,right,straight
  # same midpoint would be same segment
  #
  #                               *
  #                               |
  #  2-seg walks   *--*--*     *--*
  #                               \
  #                \ / \        \ /


  # n=5   R--*        *--L    *--L
  #       |              |       |
  #       R--L        R--L       R--L
  #          |        |             |
  #       O--L     O--L          O--L
  #      DDUU      DUDD        DDUD

  # zz(n,b) = my(z=1); for(i=0,n-2, print("i="i" "bittest(b,i)); z=1+z*I*(-1)^bittest(b,i)); z;
  # GP-Test  zz(2,0) == 1+I
  # GP-Test  zz(2,1) == 1-I

  # a(n) = if(n<2,1, sum(b=0,2^(n-2)-1, my(z=1); for(i=0,n-2, z=1+z*I*(-1)^bittest(b,i)); norm(z)==n));
  # vector(20,n,n--; c(n))

  # Knuth:
  # A = not overlapping, but not filling
  # B = not overlapping, filling, but not perfect
  # C = not overlapping, filling, perfect
  # A = [0,1,1,2,7,10,15,33,45,93,186,300,530,825,1561,2722,4685,7419,13563]
  # B = [0,0,1,0,0, 0, 0, 2, 2, 0,  0,  2,  0,  0,   8,   0,   4,   0,   12]
  # C = [1,0,1,3,0, 0, 6, 3,20, 0,  0, 29,  0,  0,  56, 101, 108,   0,  392]
  # A+B+C
  #      1,1,3,5,7,10,21,38,67,93,186,331,530,825,1625,2823,4797,7419,13967

  #
  # A+B+C == [ 1,1,3,5,7,10,21,38,67,93,186,331,530,825,1625,2823,4797,7419,13967 ]
  # A == [ 0,1,1,2,7,10,15,33,45,93,186,300,530,825,1561,2722,4685,7419,13563 ]
  # B == [ 0,0,1,0,0,0,0,2,2,0,0,2,0,0,8,0,4,0,12 ]
  # C == [ 1,0,1,3,0,0,6,3,20,0,0,29,0,0,56,101,108,0,392 ]

  # n= 6 0010 DDUDD end -1, 1 norm  2 overlap=0 doverlap=0   0,0 1,0 1,1 0,1 0,2 -1,2 -1,1
  #  L---L
  #  |   |
  #  *   R---L
  #          |
  #      O---L

  # my(z=1+I); z + z*I - 1 - I == -1 + I
  # my(z=2); z + z*I - 1 - I == 1 + I

  my $verbose = 0;
  foreach my $n (
                 $verbose ? (2 .. 4) : (2 .. 20),
                 # 5,
                 # 2 .. 50
                ) {
    my $count_all = 0;
    my $count_base_overlap = 0;
    my $count_not_base_overlap = 0;
    my $count_DDD_overlap = 0;
    my $count_not_DDD_overlap = 0;
    my $count_norm = 0;
    my $count_not_norm = 0;
    my $count_A = 0;
    my $count_B = 0;
    my $count_C = 0;
    foreach my $Sint (0 .. Sint_max($n)) {
      $count_all++;
      my $points = Sint_to_points($Sint,$n);

      my $is_overlap = points_any_overlap($points);
      if ($is_overlap) { $count_base_overlap++; }
      else { $count_not_base_overlap++; }

      my $is_norm = points_is_norm_n($points,$n);
      if ($is_norm) { $count_norm++; }
      else { $count_not_norm++; }
      my ($x,$y) = @{$points->[-1]};
      my $norm = points_end_norm($points);

      my $DDD_points = points_times_DDD($points);
      my $is_DDD_overlap = points_any_overlap($DDD_points);
      if ($is_DDD_overlap) { $count_DDD_overlap++; }
      else { $count_not_DDD_overlap++; }

      # GP-Test  ('x+I*'y) * (1+I) == ('x-'y) + ('x+'y)*I
      my $px = $x-$y - 1;
      my $py = $x+$y - 1;
      my $Dlist = [ Sint_to_Dlist($Sint,$n) ];
      scalar(@$Dlist) == $n-1 or die;
      # my $perfect_points = points_times_Dlist($points, $Dlist);
      # $perfect_points = points_times_Dlist($perfect_points, $Dlist);
      my $perfect_points = $points;
      my $is_perfect = ! points_contains_seg($perfect_points,
                                             $px,$py, $px+1,$py);

      if (! $is_DDD_overlap) {
        if (! $is_norm) {  $count_A++; }
        else {
          if (! $is_perfect) { $count_B++; }
          else { $count_C++; }
        }
      }

      if ($verbose) {
        my $Dstr = join('',@$Dlist);
        printf "n=%-2d %0*b %s end %2d,%-2d norm %2d[%d] overlap=%d doverlap=%d pxy=%d,%d perf=%d   %s\n",
          $n, $n-2,
          $Sint, $Dstr,
          $x,$y, $norm, $is_norm,
          $is_overlap, $is_DDD_overlap,
          $px,$py, $is_perfect,
          points_str($points);
        if (! $is_DDD_overlap) {
          my $conf = points_to_conf($points, name => $Dstr);
          $conf->print;
          $conf->num_segs == $n or die;

          my $DDD_conf = points_to_conf($DDD_points, name => "$Dstr square");
          $DDD_conf->print;
          $DDD_conf->num_segs == 4*$n or die;

          my $perfect_conf = points_to_conf($perfect_points, name => "$Dstr perfect");
          my $contains = points_contains_seg($perfect_points,
                                             $px,$py, $px-1,$py);
          print "  contains $contains    $px,$py, ",$px-1,",$py\n";
          $perfect_conf->print;
          # $perfect_conf->num_segs == $n**3 or die;
        }
      }
    }
    print $count_C,",";
    if ($verbose) {
      print " of total $count_all\n";
    }

    $count_base_overlap + $count_not_base_overlap == 2**($n-2) or die;
  }
  exit 0;
}

sub Sint_max {
  my ($n) = @_;
  return 2**($n-2) - 1;
}

sub Sint_to_Dstr {
  my ($Sint, $n) = @_;
  return join('', Sint_to_Dlist($Sint,$n));
}
sub Sint_to_Dlist {
  my ($Sint, $n) = @_;
  my @ret = 'D';
  my $bit = 1;
  foreach (3 .. $n) {
    push @ret, ($Sint & $bit ? 'U' : 'D');
    $bit <<= 1;
  }
  return @ret;
}

sub Sint_to_points {
  my ($Sint, $n) = @_;
  $n >= 2 or die;
  if ($Sint >> ($n-2)) { die; }
  my @ret = ([0,0],[1,0],[1,1]);
  my $x = 1;
  my $y = 1;
  my $dx = 0;
  my $dy = 1;
  my $bit = 1;
  foreach (3 .. $n) {
    if ($Sint & $bit) {
      ($dx,$dy) = xy_rotate_minus90($dx,$dy);
    } else {
      ($dx,$dy) = xy_rotate_plus90($dx,$dy);
    }
    $x += $dx;
    $y += $dy;
    push @ret, ([$x,$y]);
    $bit <<= 1;
  }
  @ret == $n+1 or die;
  return \@ret;
}
CHECK {
  points_str(Sint_to_points(0,2)) eq '0,0 1,0 1,1' or die;
  points_str(Sint_to_points(0b110,5))  eq '0,0 1,0 1,1 0,1 0,2 1,2' or die;
}

sub points_any_overlap {
  my ($points) = @_;
  my %seen;
  foreach my $i (1 .. $#$points) {
    my $key = ($points->[$i-1]->[0] + $points->[$i]->[0])  # midpoint x,y
      . ' ' . ($points->[$i-1]->[1] + $points->[$i]->[1]);
    if ($seen{$key}++) {
      return 1;
    }
  }
  return 0;
}
CHECK {
  points_any_overlap([[0,0],[1,0]]) == 0 or die;
  points_any_overlap([[0,0],[1,0],[0,0]]) == 1 or die;
}

sub points_is_norm_n {
  my ($points, $n) = @_;
  return (points_end_norm($points) == $n);
}
sub points_end_norm {
  my ($points) = @_;
  my ($x,$y) = @{$points->[-1]};
  return $x*$x + $y*$y;
}

sub points_append_D {
  my ($points, $append, $turn) = @_;
  $turn //= 'D';
  $append = MyPlanar::points_rotate_plus90
    ($append,
     points_end_dir4($points) + ($turn eq 'D' ? 1 : -1)
     - points_start_dir4($append));
  ((points_end_dir4($points) + ($turn eq 'D' ? 1 : -1)) & 3)
    == points_start_dir4($append) or die;

  $append = MyPlanar::points_move($append,
                                  $points->[-1]->[0] - $append->[0]->[0],
                                  $points->[-1]->[1] - $append->[0]->[1]);

  point_str($points->[-1]) eq point_str($append->[0]) or die;
  shift @$append;

  return [ @$points, @$append ];
}

sub points_times_D {
  my ($points) = @_;
  my $rev = [ reverse @$points ];
  return points_append_D($points, $rev);
}

sub points_times_Dlist {
  my ($points, $Dlist) = @_;
  my $reverse = [ reverse @$points ];
  my $ret = $points;
  foreach my $turn (@$Dlist) {
    ($points,$reverse) = ($reverse,$points);
    $ret = points_append_D($ret, $points, $turn);
  }
  return $ret;
}

sub points_times_DDD {
  my ($points) = @_;
  my $ret = points_times_Dlist($points,['D','D','D']);

  # my $rev = [ reverse @$points ];
  # 
  # my $ret = $points;
  # $ret = points_append_D($ret, $rev);
  # $ret = points_append_D($ret, $points);
  # $ret = points_append_D($ret, $rev);
  point_str($ret->[-1]) eq '0,0' or die;
  return $ret;
}
CHECK {
  # print points_str(points_times_DDD([ [0,0], [1,0], [1,1] ]));

  #     *--*
  #     |  |
  #     *--*--*
  #        |  |
  #        *--*
  points_str(points_times_DDD([ [0,0], [1,0], [1,1] ]))
    eq '0,0 1,0 1,1 0,1 0,2 -1,2 -1,1 0,1 0,0' or die;

  # {
  #   my $path = Math::PlanePath::R5DragonCurve->new;
  #   my $points = [ map {[$path->n_to_xy($_)]} 0 .. 5 ];
  #   my $want = [ map {[$path->n_to_xy($_)]} 0 .. 20 ];
  #   points_str(points_times_DDD($points)) eq points_str($want) or die;
  # }
  # {
  #   my $path = Math::PlanePath::DragonCurve->new;
  #   my $points = [ map {[$path->n_to_xy($_)]} 0 .. 4 ];
  #   my $want = [ map {[$path->n_to_xy($_)]} 0 .. 16 ];
  #   points_str(points_times_DDD($points)) eq points_str($want) or die;
  # }
}

sub points_contains_seg {
  my ($points, $x1,$y1, $x2,$y2) = @_;
  foreach my $i (1 .. $#$points) {
    if (($points   ->[$i-1]->[0] == $x1
         && $points->[$i-1]->[1] == $y1
         && $points->[$i]->[0] == $x2
         && $points->[$i]->[1] == $y2)
        ||
        ($points   ->[$i-1]->[0] == $x2
         && $points->[$i-1]->[1] == $y2
         && $points->[$i]->[0] == $x1
         && $points->[$i]->[1] == $y1)) {
      return 1;
    }
  }
  return 0;
}

sub points_start_dir4 {
  my ($points) = @_;
  return dxdy_to_dir4($points->[1]->[0] - $points->[0]->[0],
                      $points->[1]->[1] - $points->[0]->[1]);
}
sub points_end_dir4 {
  my ($points) = @_;
  return dxdy_to_dir4($points->[-1]->[0] - $points->[-2]->[0],
                      $points->[-1]->[1] - $points->[-2]->[1]);
}
sub points_to_conf {
  my ($points, @args) = @_;
  MyDragonConf->new
    (@args,
     segs => [ map {
       MyDragon::seg_normalize($points->[$_-1]->[0]
                               . ',' .
                               $points->[$_-1]->[1]
                               . ',' .
                               dxdy_to_dir4($points->[$_]->[0] - $points->[$_-1]->[0],
                                            $points->[$_]->[1] - $points->[$_-1]->[1]))
       } 1 .. $#$points ]);
}
