#!/usr/bin/perl -w

# Copyright 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.010;
use strict;
use Carp 'croak';
use File::Spec;
use List::Util 'max','min','sum';
use Math::BaseCnv 'cnv';

use Math::PlanePath::DragonCurve 118; # v.118 for xyxy_to_n_either()
use Math::PlanePath::Base::Digits
  'bit_split_lowtohigh', 'digit_split_lowtohigh',
  'digit_join_lowtohigh', 'round_down_pow';
use Math::PlanePath::Base::Generic
  'round_nearest';

# use Math::PlanePath;
# *_divrem_mutate = \&Math::PlanePath::_divrem_mutate;
# unshift @INC, '/usr/share/perl5';

use FindBin;
use lib::abs "$FindBin::Bin/../tools";
use MyDragon ':all';
our @ipc;
$|=1;

# uncomment this to run the ### lines
# use Smart::Comments;


{
  # Recurrence Function
  # not in OEIS: 4,9, 34,35, 68,69,70,71, 136,137,138,139,140,141,142,143, 272
  # not in OEIS:             68,69,70,71, 136,137,138,139,140,141,142,143, 272

  # 4,
  # 9,
  # 34,35,
  # 68,69,70,71,
  # 136,137,138,139,140,141,142,143,
  # 272,273,274,275,276,277,278,279,280,281,282,283,284,285,286,287,
  # 544,545,546,547,548,549,550,551,552,553,554,555,556,557,558,559,560,561,562,563,564,565,566,567,568,569,570,571,572,573,574,575

  # GP-Test  34*8  == 272
  # GP-Test  34*16 == 544

  # GP-DEFINE  H(k) = my(h=logint(k,2)); if(k==1<<h, h-1, h);
  # GP-Test  vector(1024,k, 2^H(k) < k && k <= 2^(H(k)+1)) == \
  # GP-Test  vector(1024,k, 1)
  # GP-Test  vector(1024,k,k++; H(k)) == \
  # GP-Test  vector(1024,k,k++; logint(k-1,2))
  #
  # GP-DEFINE  \\ upper bound per Allouche and Bousquet-Melou
  # GP-DEFINE  Rec(k) = 44*2^H(k) - 2^H(k+1) + k - 2;
  # vector(17,k,k++; Rec(k))
  # 42,87,86,175,176,177,174,351,352,353,354,355,356,357,350,703,704
  #
  # cf Thue-Morse
  # A005942 factor complexity
  #   2,4,6,10,12,16,20,22,24,28,32,36,40,42,44,46
  # A245186 2-Abelian
  # A179867 "Recurrence function"
  #   3,9,11,21,22,41,42,43,44,81,82,83,84,85,86,87,88,161,


  my $max_len = 700;
  # $seen[$len] = hashref of str => 1 for the turns subwords str
  my @seen;
  {
     my $str = join('', map {BitAboveLowestOne($_)} 1 .. ($max_len << 6));
    # my $str = join('', map {ThueMorse($_)} 1 .. ($max_len << 8));
    foreach my $len (1 .. $max_len)  {
      foreach my $pos (0 .. length($str) - $len)  {
        $seen[$len]->{substr($str,$pos,$len)} = 1;
      }
      # if ($len >= 7) { scalar(keys %{$seen[$len]}) == 4*$len or die; }
    }
  }
  my $str_contains_all = sub {
    my ($str, $subhref) = @_;
    foreach my $substr (keys %$subhref) {
      if (index($str,$substr) < 0) {
        return 0;
      }
    }
    return 1;
  };
  my $all_contain_all = sub {
    my ($href, $subhref) = @_;
    foreach my $str (keys %$href) {
      $str_contains_all->($str, $subhref) or return 0;
    }
    return 1;
  };

 N: foreach my $n (1 .. $max_len) {
    foreach my $R ($n .. $max_len) {
      if ($all_contain_all->($seen[$R],$seen[$n])) {
        print "$R,";
        next N;
      }
    }
    print "\nnot found at n=$n\n";
    last;
  }
  print "\n";
  exit 0;
}
{
  # count distinct turn sub-sequences
  # 2,4,8,12,18,23,28,32,36,40,44,48,52,56,60,64,
  # A337120

  # length 4, 12 of
  #   0001
  #   0010
  #   0011
  #   0100
  #   0110
  #   0111
  #
  #   1000
  #   1001
  #   1011
  #   1100
  #   1101
  #   1110

  my $max_len = 16;
  my @seen;
  # my $str = join('', map {BitAboveLowestOne($_)} 1 .. ($max_len << 6));
  my $str = join('', map {ThueMorse($_)} 1 .. ($max_len << 6));
  foreach my $len (1 .. $max_len)  {
    foreach my $pos (0 .. length($str) - $len)  {
      $seen[$len]->{substr($str,$pos,$len)} = 1;
    }
  }
  my @counts;
  foreach my $i (1 .. $max_len)  {
    my $href = $seen[$i];
    my $count = scalar(keys %$href);
    push @counts, $count;
    print "$count,";
  }
  print "\n";

  {
    my $href = $seen[4];
    my @strings = keys %$href;
    @strings = sort @strings;
    foreach (@strings) { print $_,"\n"; }
  }
  exit 0;
}


{
  # sync positions

  my $limit=16;
  my @P = (undef,
           [1,2,3,6]);
  # foreach my $k (2 .. $limit>>1) {
  # foreach my $bit (0,1) {
  #   my $t = 2*$k + $bit;
  #   next if $t > $limit;
  #   my $Pk = $P[$k];
  #   my @new;
  #   if ($bit==0) {
  #     @new = map {(2*$_-1, 2*$_)} @{$P[$k]};
  #   } else {
  #     @new = ((map {2*$_-1} @{$P[$k]}),
  #             (map {2*$_  } @{$P[$k+1]}));
  #   }
  #   $P[$t] = [sort {$a<=>$b} List::Util::uniqnum(@new)];
  # }
  foreach my $k (2 .. $limit) {
    my $H = 0;
    until (2**$H < $k && $k <= 2**($H+1)) { $H++; }
    my @new = ((1 .. 3*2**$H),
               (5*2**$H - $k + 1 .. 4*2**$H),
               (6*2**$H - $k + 1 .. 6*2**$H),
               (12*2**$H - $k + 1 .. 11*2**$H),
               (13*2**$H - $k + 1 .. 12*2**$H));
    $P[$k] = [sort {$a<=>$b} List::Util::uniqnum(@new)];
    my $len = scalar(@{$P[$k]});
    my $newlen = scalar(@new);
    print "k=$k len $newlen vs $len : ",join(',',@{$P[$k]}),"\n";
  }
  foreach my $k (1 .. $#P) {
  }
  exit 0;
}
{
  # turn subword complexity by shape of resulting curve part
  # cf
  # points in order same as paperfolding word (turns determined by points)
  # A337120
  # 0,2,4,8,12,18,23,28,32,36,40,44,48,52,56,60,64,68,72,76,80,84,88,92,96,100,104,108,112,116,120,124,
  #
  # points sorted, same with two exceptions, maybe:
  # 0,2,4,8,12,18,22,27,32,36,40,44,48,52,56,60,64,68,72,76,80,84,88,92,96,100,104,108,112,116,120,124,
  #               ^^ ^^

  require Math::NumSeq::PlanePathTurn;
  my $max_len = 256;

  # $seen[$len] = hashref of str => 1 for the turns subwords str
  my @seen;
  my $str = join('', map {BitAboveLowestOne($_)} 1 .. ($max_len << 6));
  foreach my $len (1 .. $max_len)  {
    foreach my $pos (0 .. length($str) - $len)  {
      $seen[$len]->{substr($str,$pos,$len)} = 1;
    }
  }
  my @search;
  foreach my $i (0 .. $max_len-1)  {
    my $href = $seen[$i];
    my $count = 0;
    my %seen_shape;
    foreach my $turn_str (keys %$href) {
      my @turns = split '', $turn_str;
      my $shape = turns_to_shape(\@turns);
      ### @turns
      ### $shape
      next if $seen_shape{$shape}++;
      $count++;
    }
    print "$count,";
    push @search, $count;
  }
  print "\n";

  foreach my $i (1 .. $#search) {
    if ($search[$i] != 4*$i) {
      print "$i  $search[$i]\n";
    }
  }

  print "\n";
  require Math::OEIS::Grep;
  Math::OEIS::Grep->search(array => \@search, verbose => 1);
  exit 0;

  sub turns_to_shape {
    my ($aref) = @_;
    my @points = ('0,0');
    my ($dx,$dy) = (1,0);
    my ($x,$y) = ($dx,$dy);
    foreach my $turn (@$aref) {
      push @points, "$x,$y";
      if ($turn) {
        ($dx,$dy) = xy_rotate_plus90($dx,$dy);  # left
      } else {
        ($dx,$dy) = xy_rotate_minus90($dx,$dy);  # right
      }
      $x += $dx;
      $y += $dy;
    }
    push @points, "$x,$y";
    @points = sort @points;
    return join(' ',@points);
  }
}
{
  # Abelian subwords count, by Foma

  # a(4n) = a(2n),                       n00 -> n0              collapse low 0s
  # a(4n+2) = a(2n+1) + 1,               n10 -> n1     +1       low run of 0s
  # a(16n+1) = a(8n+1),                n0001 -> n001            collapse 0s
  # a(16n+{3,7,9,13}) = a(2n+1) + 2,   n0011 -> n1    +2
  #                                    n0111 -> n1    +2
  #                                    n1001 -> n1    +2
  #                                    n1101 -> n1    +2
  # a(16n+5) = a(4n+1) + 1,            n0101 -> n01   +1
  # a(16n+11) = a(4n+3) + 1,           n1011 -> n11   +1       run 0s
  # a(16n+15) = a(2n+2) + 1.           n1111 -> n0    +1       3 1s
  #                                             ^+carry
  # Minimized:
  # 0[1]111  +2
  #   (10)1  +2 each
  #     011  +2
  #     001  +2
  #  1[0]00  +2
  #     110  +2
  #   (01)0  +2 each
  #
  my @table = (['00',   '0',   'e'],
               ['10',   '1',   'P'],
               ['0001', '001', 'e'],
               ['0011', 's',   'P'], ['s', '1',  'P'],
               ['0111', 't',   'P'], ['t', '1',  'P'],
               ['1001', 'u',   'P'], ['u', '1',  'P'],
               ['1101', 'v',   'P'], ['v', '1',  'P'],
               ['0101', 'w',   'P'], ['w', '01', 'P'],
               ['1011', 'x',   'P'], ['x', '11', 'P'],
               ['1111', 'c0',  'P']);
  my %lows = map {$_->[0] => 1} @table;
  my %table;
  {
    my %seen;
    foreach my $elem (@table) {
      my ($from,$to,$out) = @$elem;
      $table{$from}->{'e'} = [$to,$out];
      $seen{$from} = 1;
      unless ($to =~ /c/) {
        $table{"c$from"}->{'e'} = ["c$to",$out];
        $seen{"c$from"} = 1;
      }
    }
  N: foreach my $n (1..31) {
      my $from = sprintf '%b',$n;
      $from = substr($from,1);
      foreach my $elem (@table) {
        my $str = $elem->[0];
        next N if $from =~ /$str$/
      }
      unless (exists $table{"c$from"}) {
        ### carry: $from
        $table{"c$from"}->{0} = [ "1$from", 'e'];
        $table{"c$from"}->{1} = ["c0$from", 'e'];
      }
      unless (exists $table{$from}) {
        ### plain: $from
        $table{$from}->{0} = [ "0$from", 'e'];
        $table{$from}->{1} = [ "1$from", 'e'];
      }
    }
  }
  {
    for (;;) {
      my %reachable = ('' => 1);
      while (my ($from,$href) = each %table) {
        while (my ($symbol,$elem) = each %$href) {
          my ($to,$out) = @$elem;
          ### reachable: $to
          $reachable{$to} = 1;
        }
      }
      my $more = 0;
      foreach my $from (keys %table) {
        if (! $reachable{$from}) {
          ### delete: $from
          delete $table{$from};
          $more = 1;
        }
      }
      last unless $more;
    }
  }
  my $upto_state = 0;
  my %str_to_state;
  my @state_to_str;
  foreach my $from (sort keys %table) {
    print "\"$from\"\n";
    $str_to_state{$from} = $upto_state;
    $state_to_str[$upto_state] = $from;
    $upto_state++;
    foreach my $symbol ('e',0,1) {
      my $elem = $table{$from}->{$symbol} // next;
      my ($to,$out) = @$elem;
      print "  $symbol -> \"$to\" out $out\n";
    }
  }
  $str_to_state{''} == 0 or die;
  {
    my %symbol_to_att = (0 => 'z',
                         1 => 1,
                         2 => 2,
                         P => 'P',
                         e => '@0@');
    my $att_filename = '/tmp/Abelian.att';
    open my $fh, '>', $att_filename or die;
    foreach my $state (0 .. $#state_to_str) {
      my $from = $state_to_str[$state];
      print "$state str \"$from\"\n";
      if (defined $table{$from}->{'e'}
          && (defined $table{$from}->{'0'}
              || defined $table{$from}->{'1'})) {
        die;
      }
      foreach my $symbol ('e',0,1) {
        my $elem = $table{$from}->{$symbol} // next;
        ### $from
        ### $symbol
        ### $elem
        my ($to,$out) = @$elem;
        my $to_state = $str_to_state{$to};
        my $line = "$state\t$to_state\t$symbol_to_att{$symbol}\t$symbol_to_att{$out}\n";
        print $fh $line;
        # print $line;
      }
    }
    foreach my $accepting (# '', '0',
                           '1') {
      my $state = $str_to_state{$accepting} // die;
      print $fh "$state\n";
    }
    close $fh or die;
  }

  require Math::OEIS::Stripped;
  my @want = (0,  # OFFSET=1
              Math::OEIS::Stripped->anum_to_values('A214613'));
  # 2,3,4,3,4,5,4,3,4,5,6,5,4,5,4,3,4,5,6,5,6,7,6,5

  my $funcinfo = sub {
    my ($n) = @_;
    my $state = '';
    my $ret = 0;
    my $n2 = reverse(sprintf '%b',$n);
    ### func: "$n  $n2"
    my $pos = 0;
    while ($pos < length($n2)) {
      ### at: "state=\"$state\" pos=$pos bit ".substr($n2,$pos,1)." ret $ret"
      my $elem = $table{$state}->{'e'}
        // $table{$state}->{substr($n2,$pos++,1)};
      my ($to,$out) = @$elem;
      $state = $to;
      if ($out eq 'P') {
        ### p ...
        $ret++;
      }
    }
    ### end loop: "state=\"$state\" ret $ret"
    my %state_to_final = (
                          ''     => -2,
                          '0'    => -2,
                          '10'   => 1,
                          '11'   => 2,
                          '101'  => 2,
                          '111'  => 2,
                          '1001' => 2,
                          '1011' => 4,
                          '1101' => 2,
                          '1111' => 2,
                          'c00'  => 1,
                         );
    $ret += 2;
    $ret += ($state_to_final{$state} || 0);
    ### final: "state=\"$state\" ret $ret"
    return ($state,$ret);
  };
  my $func = sub {
    my ($n) = @_;
    [$funcinfo->($n)]->[1];
  };
  my %state_to_dd;
  foreach my $n (0.. $#want) {
    $func->($n) == $want[$n] or die "oops n=$n"; }
  ### %state_to_dd

  foreach my $delta (1, 3) {
    my $count = 0;
    foreach my $n (1024 .. 2048,
                  ) {
      my ($state,$got) = $funcinfo->($n);
      my $n2 = sprintf '%b', $n;
      my $dir = dir($n);
      my $dd = $dir - $got;
      if ($got == $dir + $delta  && $dir == 8) {
        print "n=$n  [$n2  $got $dir  \"$state\"\n";
        last if $count++ > 150;
      }
    }
    print "\n";
  }
  # my $want = $want[$n] // 'xx';
  # my $diff = ($n <= $#want && $got != $want ? "  **** ".($got-$want) : '');
  # my $n2 = sprintf '%b', $n;
  # my $dir = dir($n) + 2;
  # my $dd = $dir - $got;
  # $state_to_dd{$state}->{$dd} = 1;
  # print "n=$n [$n2]  $got want $want$diff  $dd  \"$state\"\n";
  exit;

  unlink '/tmp/xyz.att';
  unlink '/tmp/xyz.dot';
  run_foma(
           "read att /tmp/Abelian.att"."\n",
           "minimize net\n",
           "define Ab\n",
           # "regex [ Ab ] .o. [ P -> t  ];\n"
           # "reverse net\n"
           # "define dir [z|1]* .o. [ [ 1 1 | z z ] -> 0 ,, [1 z|z 1] -> P ,, 1 -> P || _ .#.];\n"
           # "define dir [z|1]* .o. [ 1 1* -> P ,, z z* -> P || _ 1 ];\n"
           # "define Gray [z|1]* .o. [ [ 1 1 | z z ] -> S ,, [ 1 z | z 1 ] -> P ,, 1 -> P || _ .#. ];\n"
           "define one    [ [ P P P P P P         ] : [ W ] ];\n",
           "define two    [ [ P P P P P P P       ] : [ W ] ];\n",
           "define three  [ [ P P P P P P P P     ] : [ W ] ];\n",
           "define four   [ [ P P P P P P P P P   ] : [ W ] ];\n",
           "define five   [ [ P P P P P P P P P P ] : [ W ] ];\n",
           "define Runs [z|1]* .o. [ 1 -> 1 P || _ [z | .#.] ,, z -> z P || _ 1 ] .o. [ [1|z] -> 0 ];\n",
           "regex [Ab .o. three].u & [Runs .o. four].u;\n",
           # "reverse net\n",
           "print upper-words\n",
           "write att /tmp/xyz.att"."\n",
           # "regex [x y z] : 0;\n"
           # "print dot >/tmp/Abelian.dot"."\n"
          );
  run_foma("read att /tmp/xyz.att"."\n",
           "print dot >/tmp/xyz.dot"."\n");
  require MyGraphs;
  MyGraphs::graphviz_view_file('/tmp/xyz.dot');
  exit 0;

  sub run_foma {
    my $str = join('',@_);
    require File::Slurp;
    require IPC::Run;
    File::Slurp::write_file('/tmp/tempfile.foma',$str);
    IPC::Run::run(['foma','-p','-f','/tmp/tempfile.foma'],
                  '<','/dev/null') or die;
  }
  sub run_hfst {
    my $str = join('',@_);
    require File::Slurp;
    require IPC::Run;
    File::Slurp::write_file('/tmp/tempfile.foma',$str);
    IPC::Run::run(['hfst-xfst','-F','/tmp/tempfile.foma'],
                  '<','/dev/null') or die;
  }
}
{
  # count distinct turn subwords, by DFAs
  # 2,4,8,12,18,23,28,32,36,40,44,48,52,56,60,64,68,72,76,80,84,88,92,96,100,104,108,112,116,120,124,128,132,136,140,144,148,152,156,160,164,168,172,176,180,184,188,192,196,200,204,208,212,216,220,224,228,232,236,
  # A337120
  # initial exceptions [2, 4, 8, 12, 18, 23]
  # 4*n
  # = 2*A005943 GRS subwords

  my @bit_to_flat = (TurnLpred_flat(),
                     TurnRpred_flat());
  my @pending = map {$bit_to_flat[$_]->clone->MyFLAT::set_name($_)} 0,1;
  print scalar(@pending),",";
  foreach my $len (2 .. 100) {
    my @new_pending;
    foreach my $f (@pending) {
      my $f_inc = $f->MyFLAT::digits_increment(add => 1);
      foreach my $bit (0,1) {
        my $new_f = $f_inc->intersect($bit_to_flat[$bit]);
        if (! $new_f->is_empty) {
          my $name = $f->{'name'};
          $new_f = $new_f->MyFLAT::set_name("$bit$name")
            ->MyFLAT::minimize (verbose => 0);
          push @new_pending, $new_f;
        }
      }
    }
    @pending = @new_pending;
    print scalar(@pending),",";
  }
  if (@pending < 16) {
    my @strings = map {$_->{'name'}} @pending;
    @strings = sort @strings;
    foreach (@strings) { print $_,"\n"; }
  }
  exit 0;

  # 1100

  # 0001
  # 0011
  # 0101
  # 0111
  # 1001
  # 1011
  # 1101
  # 1111

}


sub ThueMorse {
  my ($n) = @_;
  my $ret = 0;
  while ($n) { $ret ^= $n&1; $n >>= 1; }
  return $ret;
}
