#!/usr/bin/perl -w

# Copyright 2016, 2017, 2018, 2019, 2021 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.

use strict;
use 5.006;
use Carp 'croak';
use FLAT;
use FLAT::Regex;
use FLAT::Regex::WithExtraOps;
use FLAT::DFA;
use FLAT::PFA;
use Math::BaseCnv;
use Test::More;

use FindBin;
use lib $FindBin::Bin;
use MyFLAT ':all';

use MyTestHelpers;
BEGIN { MyTestHelpers::nowarnings(); }

plan tests => 438;

# uncomment this to run the ### lines
# use Smart::Comments;


#------------------------------------------------------------------------------
# write_att_file()

{
  my $fa = FLAT::NFA->new;
  ok ($fa->is_empty);
  is ($fa->MyFLAT::finite_max_length, -1,
     'finite_max_length() of empty');
  $fa->add_states(3);
  $fa->set_starting(0);
  $fa->set_accepting(2);
  $fa->add_transition(0,1, '');
  $fa->add_transition(1,2, '');
  $fa->add_transition(2,2, 'x');
  ok (join(",", sort $fa->epsilon_closure(0)), '1,2,3');

  my $want_str = <<"HERE";
0\t1\t\@0\@\t\@0\@
1\t2\t\@0\@\t\@0\@
2\t2\tx\tx
2
HERE
  { require IO::String;
    my $fh = IO::String->new(my $got_str);
    $fa->MyFLAT::write_att_fh($fh);
    is ($got_str, $want_str, 'write_att_fh()');
  }

  ### $want_str
  { require IO::String;
    my $fh = IO::String->new($want_str);
    my $read = FLAT::DFA->MyFLAT::read_att_fh($fh);
    # $read->MyFLAT::view;
    ok ($fa->as_dfa->equals($read->as_dfa));
  }
}


#------------------------------------------------------------------------------
# digits_multiply()

{
  my $fa = FLAT::Regex->new('123')->as_dfa->MyFLAT::minimize;
  my $new_fa = $fa->MyFLAT::digits_multiply (mul => 5);
  # $fa->MyFLAT::view;
  # $new_fa->MyFLAT::view;
  ok ($new_fa->equals(FLAT::Regex->new('60')->as_dfa));
}


#------------------------------------------------------------------------------
# transmute()

{
  my $transmute_initial_zeros = sub {
    my ($state,$symbol) = @_;
    unless ($state) {
      if ($symbol eq '1') { $state = 'done'; }   # stop when found 1
      else { $symbol = 1; } # low 0 -> 1
    }
    return ($state,$symbol);
  };
  {
    my $fa = FLAT::Regex->new('000110011')->as_dfa->MyFLAT::minimize;
    my $new_fa = $fa->MyFLAT::transmute(func => $transmute_initial_zeros);
    # $fa->MyFLAT::view;
    # $new_fa->MyFLAT::view;
    ok ($new_fa->equals(FLAT::Regex->new('111110011')->as_dfa));
  }
  {
    my $fa = FLAT::Regex->new('0001100110')->as_dfa->MyFLAT::minimize;
    my $new_fa = $fa->MyFLAT::transmute(func => $transmute_initial_zeros,
                                        direction => 'reverse');
    ok ($new_fa->equals(FLAT::Regex->new('0001100111')->as_dfa));
  }
}


#------------------------------------------------------------------------------
# perl_regexp_to_flat_regex()

is (MyFLAT::perl_regexp_to_flat_regex(''), '');
is (MyFLAT::perl_regexp_to_flat_regex('01'), '01');
is (MyFLAT::perl_regexp_to_flat_regex('[01]'), '(0|1)');
is (MyFLAT::perl_regexp_to_flat_regex('1+'), '11*');
is (MyFLAT::perl_regexp_to_flat_regex('(12)+'), '(12)(12)*');
is (MyFLAT::perl_regexp_to_flat_regex('1?'), '(1|[])');
is (MyFLAT::perl_regexp_to_flat_regex('(12)?'), '((12)|[])');

foreach my $re (qr/1+/,
               ) {
  my $flat_str = MyFLAT::perl_regexp_to_flat_regex($re);
  diag "Perl ", $re;
  diag "FLAT ", $flat_str;
  my $fa = FLAT::Regex->new($flat_str);

  my $bad = 0;
  foreach my $k (0 .. 10) {
    foreach my $n (0 .. 2**$k-1) {
      my $str = ($k==0 ? '' : sprintf '%0*b', $k, $n);
      my $by_re = ($str =~ /^$re$/ ? 1 : 0);
      my $by_fa = $fa->contains($str) ? 1 : 0;
      unless ($by_fa == $by_re) {
        print "n=$n str $str  perl $by_re flat $by_fa\n";
        die if $bad++ > 10;
      }
    }
  }
  is($bad, 0, "$re is $flat_str");
}


#------------------------------------------------------------------------------
# is_finite() and epsilon

{
  my $fa = FLAT::NFA->new;
  ok ($fa->is_empty);
  is ($fa->MyFLAT::finite_max_length, -1);
  $fa->add_states(4);
  $fa->set_starting(0);
  $fa->set_accepting(3);
  $fa->add_transition(0,1, '');
  $fa->add_transition(1,2, '');
  $fa->add_transition(2,3, '');
  $fa->add_transition(3,3, 'x');
  ok (! $fa->is_finite);
  ok ($fa->contains(''));
  ok ($fa->contains('x'));
  ok ($fa->contains('xxxxx'));
  # $fa->MyFLAT::view;
}


#------------------------------------------------------------------------------
# finite_max_length()

{
  my $fa = FLAT::NFA->new;
  ok ($fa->is_empty);
  is ($fa->MyFLAT::finite_max_length, -1);

  $fa->add_states(4);
  is ($fa->MyFLAT::finite_max_length, -1);

  $fa->set_starting(0);
  ok ($fa->is_empty);
  is ($fa->MyFLAT::finite_max_length, -1);

  $fa->set_accepting(0);
  ok (! $fa->is_empty);
  is ($fa->MyFLAT::finite_max_length, 0);

  $fa->add_transition(0,1, 9);
  $fa->add_transition(1,2, 8);
  $fa->add_transition(2,3, 7);
  $fa->set_accepting(1);
  is ($fa->MyFLAT::finite_max_length, 1);

  $fa->set_accepting(3);
  is ($fa->MyFLAT::finite_max_length, 3);
}



#------------------------------------------------------------------------------
# ExtraOps prefix

#------------------------------------------------------------------------------
# FLAT::Regex MyFLAT_skip_initial()

{
  my $fa = FLAT::Regex->new('#');
  ok ($fa->is_empty, 'FLAT::Regex # = null, no matches at all');
  ok (! $fa->op->has_nonempty_string(), 'FLAT::Regex # = null, no nonempties');
  ok (! $fa->contains(''), 'FLAT::Regex # = null, no matches at all');
}
{
  my $fa = FLAT::Regex->new('[]');
  ok (! $fa->is_empty, 'FLAT::Regex [] matches empty string');
  ok (! $fa->op->has_nonempty_string());
  ok ($fa->contains(''), 'FLAT::Regex [] matches empty string');
}
{
  my $fa = FLAT::Regex->new('a');
  ok (! $fa->is_empty);
  ok ($fa->op->has_nonempty_string());
  ok (! $fa->contains(''));
  ok ($fa->contains('a'));
}
foreach my $elem (
                  # star
                  ['a*',          'a*'],
                  ['([] a [])*',  'a*'],
                  ['([] [] [])*', '[]'],
                  ['([])*',       '[]'],
                  ['(ab)*', 'b(ab)*'],   # [], ab, abab -> b, bab
                  ['(abcd)*', 'bcd(abcd)*'],

                  # alt
                  ['# | []',           '#'],
                  ['# | #',            '#'],
                  ['[] | [] | [] | a', '[]'],
                  ['#  | [] | [] | a', '[]'],
                  ['[] | [] | []',     '#'],
                  ['aab | cab | cccd', 'ab | ab | ccd'],
                  ['[] | cab | cccd',  'ab | ccd'],
                  ['# | cab | cccd',   'ab | ccd'],

                  # atomic
                  ['[foo]', '[]'],
                  ['a',     '[]'],
                  ['[]',    '#'],

                  # concat
                  ['abc', 'bc'],
                  ['[] abc', 'bc'],
                  ['[] a [] b [] c []', 'bc'],

                 ) {
  my ($start_regex, $want_regex) = @$elem;
  my $want = FLAT::Regex->new($want_regex);
  my $start = FLAT::Regex->new($start_regex);
  {
    my $got = $start->MyFLAT_skip_initial;
    isa_ok($got, 'FLAT::Regex');
    my $want_dfa = $want->as_nfa->as_dfa;
    my $got_dfa = $got->as_nfa->as_dfa;
    ok ($got_dfa->equals($want_dfa),
        "skip_initial start \"$start_regex\" want \"$want_regex\" got \"$got\"");
  }
  {
    my $got = $start->reverse->MyFLAT_skip_final->reverse;
    my $want_dfa = $want->as_nfa->as_dfa;
    my $got_dfa = $got->as_nfa->as_dfa;
    ok ($got_dfa->equals($want_dfa),
        "skip_final start \"$start_regex\" want \"$want_regex\" got \"$got\"");
  }
}
# MyFLAT::FLAT_show_acyclics($got, 5);
# MyFLAT::FLAT_show_acyclics($want, 5);


#------------------------------------------------------------------------------
# skip_initial()

{
  my $got = FLAT::Regex->new('aab | cab | cccd')->as_nfa
    ->MyFLAT::skip_initial->as_dfa;
  my $want = FLAT::Regex->new('ab | ab | ccd')->as_dfa;
  ok ($got->equals($want));

  # MyFLAT::FLAT_show_acyclics($got, 5);
  # MyFLAT::FLAT_show_acyclics($want, 5);
  # $want->MyFLAT::view;
}


#------------------------------------------------------------------------------
# FLAT::Regex MyFLAT_suffix()

foreach my $elem (
                  # atomic
                  [0, '[foo]', '[] | [foo]'],
                  [0, 'a',     '[] | a'],
                  [0, '[]',    '[]'],
                  [0, '#',     '#'],
                  [1, '[foo]', '[]'],
                  [1, 'a',     '[]'],
                  [1, '[]',    '#'],
                  [1, '#',     '#'],

                  # concat
                  [0, 'ab', 'ab | b | []'],
                  [1, 'ab',      'b | []'],
                  [0, '(a #)',     '#'],
                  [1, '(a #)',     '#'],
                  [0, '(# a)',     '#'],
                  [1, '(# a)',     '#'],
                  [0, 'abc', 'abc | bc | c | []'],
                  [1, 'abc',       'bc | c | []'],
                  [0, 'ab [] c', 'abc | bc | c | []'],
                  [1, 'ab [] c',       'bc | c | []'],
                  [0, 'abc []', 'abc | bc | c | []'],
                  [1, 'abc []',       'bc | c | []'],
                  [0, '(#)',     '#'],
                  [1, '(#)',     '#'],

                  # star
                  [0, 'a*',          'a*'],
                  [1, 'a*',          'a*'],
                  [0, '[]*',         '[]'],
                  [1, '[]*',         '#'],
                  [0, '([] a [])*',  'a*'],
                  [1, '([] a [])*',  'a*'],
                  [0, '([] [] [])*', '[]'],
                  [1, '([] [] [])*', '#'],

                  [0, '([])*',         '[]'],
                  [1, '([])*',         '#'],

                 ) {
  my ($proper, $start_regex, $want_regex) = @$elem;
  my $want = FLAT::Regex->new($want_regex);
  my $start = FLAT::Regex->new($start_regex);
  my $got = $start->MyFLAT_suffix($proper);
  isa_ok($got, 'FLAT::Regex');
  my $want_dfa = $want->as_nfa->as_dfa;
  my $got_dfa = $got->as_nfa->as_dfa;
  ok ($got_dfa->equals($want_dfa),
      "suffix proper=$proper start \"$start_regex\" want \"$want_regex\" got \"$got\"");

  # MyFLAT::FLAT_show_acyclics($got_dfa, 5);
  # MyFLAT::FLAT_show_acyclics($want_dfa, 5);
  # MyFLAT::FLAT_check_is_equal($got_dfa,$want_dfa);
}


#------------------------------------------------------------------------------
# FLAT::Regex MyFLAT_prefix()

{
  my $fa = FLAT::Regex->new('(#)');
  ok (! $fa->op->has_nonempty_string());
  ok ($fa->is_empty);
  ok (! $fa->contains(''));
}
{
  my $fa = FLAT::Regex->new('(a* #)');
  ok ($fa->is_empty);
  ok (! $fa->contains(''));
}
{
  my $fa = FLAT::Regex->new('(a #)');
  ok ($fa->is_empty);
  ok (! $fa->contains(''));
}

foreach my $elem (
                  # These two provoke warnings in FLAT 0.9.  Are they right?
                  # ['a & #', '[]'],
                  # ['# & a', '[]'],
                  # FIXME: Is this supposed to be right?
                  # ['# & #', '#'],

                  # shuffles, successive symbol from each, in any way
                  ['a & []', 'a'],
                  ['[] & abc', 'abc'],
                  ['a & b', 'ab | ba'],
                  ['ab & c', 'abc | acb | cab'],
                  ['ab & cd', 'abcd | acbd | acdb | cabd | cadb | cdab'],
                 ) {
  my ($start_regex, $want_regex) = @$elem;
  diag "FLAT::Regex MyFLAT_prefix(): ",$start_regex;
  my $want = FLAT::Regex->new($want_regex)->as_dfa;
  isa_ok($want, 'FLAT::DFA');

  my $got_regex = FLAT::Regex::WithExtraOps->new($start_regex);
  my $got_dfa = $got_regex->as_pfa->as_nfa->as_dfa;
  isa_ok($got_dfa, 'FLAT::DFA');

  ok ($got_dfa->equals($want), "shuffle $start_regex want $want_regex got $got_regex");
  MyFLAT::FLAT_check_is_equal($got_dfa,$want);
}
foreach my $elem (
                  # shuffle
                  [0, 'a & b',   'ab | ba | a | b | []'],
                  [0, 'ab & c',  'abc | acb | cab | ab|ac|ca | a|c | []'],
                 ) {
  my ($proper, $start_regex, $want_regex) = @$elem;
  my $want = FLAT::Regex->new($want_regex);
  my $want_dfa = $want->as_nfa->as_dfa;
  my $start = FLAT::Regex::WithExtraOps->new($start_regex);
  {
    my $got = $start->MyFLAT_prefix($proper);
    ### $start
    ### $want
    ### $got
    isa_ok($got, 'FLAT::Regex');
    my $got_dfa = $got->as_pfa->as_nfa->as_dfa;
    ok ($got_dfa->equals($want_dfa),
        "prefix proper=$proper start \"$start_regex\" want \"$want_regex\" got \"$got\"");
  }

  # MyFLAT::FLAT_show_acyclics($got_dfa, 5);
  # MyFLAT::FLAT_show_acyclics($want_dfa, 5);
  # MyFLAT::FLAT_check_is_equal($got_dfa,$want_dfa);
}


foreach my $elem (
                  # concat
                  [0, '(a #)',     '#'],
                  [1, '(a #)',     '#'],
                  [0, 'abcde', '[] | a | ab | abc | abcd | abcde'],
                  [0, 'abc', '[] | a | ab | abc'],
                  [1, 'abc', '[] | a | ab'],
                  [0, 'abc []', '[] | a | ab | abc'],
                  [1, 'abc []', '[] | a | ab'],
                  [0, 'ab [] c', '[] | a | ab | abc'],
                  [1, 'ab [] c', '[] | a | ab'],
                  [0, '(#)',     '#'],
                  [1, '(#)',     '#'],

                  # atomic
                  [0, '[foo]', '[] | [foo]'],
                  [0, 'a',     '[] | a'],
                  [0, '[]',    '[]'],
                  [0, '#',     '#'],
                  [1, '[foo]', '[]'],
                  [1, 'a',     '[]'],
                  [1, '[]',    '#'],
                  [1, '#',     '#'],

                  # star
                  [0, 'a*',          'a*'],
                  [1, 'a*',          'a*'],
                  [0, '[]*',         '[]'],
                  [1, '[]*',         '#'],
                  [0, '([] a [])*',  'a*'],
                  [1, '([] a [])*',  'a*'],
                  [0, '([] [] [])*', '[]'],
                  [1, '([] [] [])*', '#'],

                  [0, '([])*',         '[]'],
                  [1, '([])*',         '#'],

                  # ['(ab)*', 'b(ab)*'],   # [], ab, abab -> b, bab
                  #
                  # # alt
                  # ['# | []',           '#'],
                  # ['# | #',            '#'],
                  # ['[] | [] | [] | a', '[]'],
                  # ['#  | [] | [] | a', '[]'],
                  # ['[] | [] | []',     '#'],
                  # ['aab | cab | cccd', 'ab | ab | ccd'],
                  # ['[] | cab | cccd',  'ab | ccd'],
                  # ['# | cab | cccd',   'ab | ccd'],

                 ) {
  my ($proper, $start_regex, $want_regex) = @$elem;
  my $want = FLAT::Regex->new($want_regex);
  my $want_dfa = $want->as_nfa->as_dfa;
  my $start = FLAT::Regex->new($start_regex);
  {
    my $start_dfa = $start->as_nfa->as_dfa;
    my $got_dfa = $start_dfa->MyFLAT::prefix($proper);
    isa_ok($got_dfa, 'FLAT::DFA');
    ok ($got_dfa->equals($want_dfa),
        "DFA prefix proper=$proper start \"$start_regex\"");
  }
  {
    my $got = $start->MyFLAT_prefix($proper);
    ### $start
    ### $want
    ### $got
    isa_ok($got, 'FLAT::Regex');
    my $got_dfa = $got->as_nfa->as_dfa;
    ok ($got_dfa->equals($want_dfa),
        "prefix proper=$proper start \"$start_regex\" want \"$want_regex\" got \"$got\"");
  }
  {
    my $got = $start->reverse->MyFLAT_suffix($proper)->reverse;
    isa_ok($got, 'FLAT::Regex');
    my $got_dfa = $got->as_nfa->as_dfa;
    ok ($got_dfa->equals($want_dfa),
        "suffix proper=$proper start \"$start_regex\" want \"$want_regex\" got \"$got\"");
  }

  # MyFLAT::FLAT_show_acyclics($got_dfa, 5);
  # MyFLAT::FLAT_show_acyclics($want_dfa, 5);
  # MyFLAT::FLAT_check_is_equal($got_dfa,$want_dfa);
}


#------------------------------------------------------------------------------
# fraction_also_nines()

{
  my $got = FLAT::Regex->new('101 | 111 | 01')->as_nfa
    ->MyFLAT::fraction_also_nines->as_dfa;
  my $want = FLAT::Regex->new('101 | 111 | 01
                             | 1001* | 1101* | 001*')->as_dfa;
  ok ($got->equals($want));
}

#------------------------------------------------------------------------------
# _divrem()

{
  my ($q,$r) = MyFLAT::_divrem(20,3);
  is($q,6);
  is($r,2);
}

#------------------------------------------------------------------------------
# digits_increment()

{
  my $got = FLAT::Regex->new('15')
    ->MyFLAT::digits_increment (radix=>10,
                                add=>-52,
                                direction=>'lowtohigh');
  ok ($got->is_empty);
}
{
  my $got  = FLAT::Regex->new('15')->as_nfa;
  # $got->MyFLAT::view;
  $got = $got->MyFLAT::digits_increment (radix=>10,
                                         add=>-51,
                                         direction=>'lowtohigh');
  # $got->MyFLAT::view;
  $got = $got->as_dfa;
  my $want = FLAT::Regex->new('00')->as_dfa;
  ok ($got->equals($want));
  # MyFLAT::FLAT_show_breadth($got, 4);
  # MyFLAT::FLAT_show_breadth($want, 4);
}

{
  my $got  = FLAT::Regex->new('11 0*')->as_dfa;
  $got = $got->MyFLAT::digits_increment (direction=>'lowtohigh');
  # $got->MyFLAT::view;
  my $want = FLAT::Regex->new('001 0*')->as_dfa;
  ok ($got->equals($want));
}

{
  my $got  = FLAT::Regex->new('1111 0*')->as_dfa
    ->MyFLAT::digits_increment (direction=>'lowtohigh');
  my $want = FLAT::Regex->new('00001 0*')->as_dfa;
  ok ($got->equals($want));
}

foreach my $n (55, 19, 1999, 89) {
  my $str        = reverse $n;
  my $str_plus_1 = reverse $n+1;
  my $got  = FLAT::Regex->new($str)->as_dfa
    ->MyFLAT::digits_increment (radix=>10, direction=>'lowtohigh');
  my $want = FLAT::Regex->new($str_plus_1)->as_dfa;
  ok ($got->equals($want));
}



{
  # 1, 11, 111, 1111, ... -> 01, 001, 0001, 00001, ...
  my $got  = FLAT::Regex->new('1*')->as_dfa
    ->MyFLAT::digits_increment (radix=>2, direction=>'lowtohigh');
  my $want = FLAT::Regex->new('0*1')->as_dfa;
  ok ($got->equals($want));

}
{
  my $got  = FLAT::Regex->new('1111')->as_dfa
    ->MyFLAT::digits_increment (radix=>2, direction=>'lowtohigh');
  my $want = FLAT::Regex->new('00001')->as_dfa;
  ok ($got->equals($want));
}


{
  foreach my $direction ('lowtohigh','hightolow') {
    foreach (1 .. 5) {
      my $len = int(rand(10))+1;
      my $radix = int(rand(5))+2;
      my @numbers = map { int(rand(100)) } 1 .. $len;
      my @strings        = map { Math::BaseCnv::cnv($_,10,$radix) } @numbers;
      my @strings_plus_1 = map { Math::BaseCnv::cnv($_+1,10,$radix) } @numbers;
      if ($direction eq 'lowtohigh') {
        @strings        = map { scalar reverse } @strings;
        @strings_plus_1 = map { scalar reverse } @strings_plus_1;
      }

      my $re        = join('|',@strings);
      my $re_plus_1 = join('|',@strings_plus_1);
      my $got  = FLAT::Regex->new($re)->as_dfa
        ->MyFLAT::digits_increment (radix=>$radix, direction=>$direction);
      my $want = FLAT::Regex->new($re_plus_1)->as_dfa;
      ### @numbers
      ### @strings
      ### @strings_plus_1
      ### $re
      ### $re_plus_1
      ok ($got->equals($want));
    }
  }
}


#------------------------------------------------------------------------------
# MyFLAT::get_eventually_accepting_info()

{
  my $fa = FLAT::DFA->new;
  my @states = $fa->add_states(2);
  $fa->add_transition($states[0], $states[1], [0,1,2]);
  $fa->add_transition($states[1], $states[1], [0,1,2]);
  $fa->set_accepting($states[1]);

  is_deeply([$fa->MyFLAT::get_non_accepting], [0]);

  {
    my %info = MyFLAT::get_eventually_accepting_info($fa);
    is ($info{'depth'}, 1);
    is_deeply ($info{'states'}, [0]);
  }

  $fa = MyFLAT::eventually_accepting($fa);
  {
    my %info = MyFLAT::get_eventually_accepting_info($fa);
    is ($info{'depth'}, 0);
    is_deeply ($info{'states'}, []);
  }
}


#------------------------------------------------------------------------------
# FLAT_count_contains()

{
  my $fa = FLAT::Regex->new('(0|1)*')->as_dfa;
    my $max_k = 10;
    is_deeply([ FLAT_count_contains($fa,$max_k) ],
              [ map {2**$_} 0 .. $max_k ]);
  }


#------------------------------------------------------------------------------
# fraction_digits()

ok(fraction_digits(0,1, radix => 2)->equals(zero_digits_flat(2)),
   "fraction 0/1 binary = zero bits");

{
  my $f = fraction_digits(1,15, radix => 2);
  ok($f->contains(''));
  ok($f->contains('0'));
  ok($f->contains('0001'));
  ok($f->contains('00010'));

  my $r = FLAT::Regex->new('(0001)* ([] | 0 | 00 | 000)')->as_dfa;
  ok ($f->equals($r), "fraction 1/15 = regex");
}

{
  my $f = fraction_digits(1,1);
  ok($f->contains(''));
  ok($f->contains('9'));
  ok($f->contains('99'));
  ok(! $f->contains('0'));
  ok(! $f->contains('8'));
}

{
  my $f = fraction_digits(99,100);
  # MyFLAT::view($f);
  ok($f->contains(''));
  ok($f->contains('9'));
  ok($f->contains('99'));
  ok($f->contains('990'));
  ok($f->contains('9900'));
  ok($f->contains('98'));
  ok($f->contains('989'));
  ok($f->contains('9899'));
  ok($f->contains('98999'));
  ok(! $f->contains('988'));
  ok(! $f->contains('991'));
  ok(! $f->contains('0'));
  ok(! $f->contains('8'));
  ok(! $f->contains('90'));
}

# {
#   my $len = 10;
#   my $bad = 0;
#   foreach my $radix (2, 3, 4, 8, 9, 10, 16) {
#     foreach my $den (1 .. 50) {
#       foreach my $num (1 .. $den) {
#         my $full_str = cnv( int($num * $radix**$len / $den), 10,$radix);
#         foreach my $i (0 .. $len) {
#           my $str = substr($full_str,0,$len);
#           my $got = fFifthsPred_flat()->contains($str) ? 1 : 0;
# 
#         # at $i==$len must have got=want as have seen enough bits,
#         # prior to then got might accept more than want
#         unless ($i==$len ? $got == $want : $got >= $want) {
#           print "fFifthsPred_flat() $num/$den N=$n $str got $got want $want\n";
#           $bad++;
#           last DEN if $bad++ > 10;
#         }
#       }
#     }
#   }
#   is($bad, 0, "fFifthsPred_flat() on fractions");
# }

#------------------------------------------------------------------------------
# one_bits_flat()

ok(one_bits_flat()->contains(''));
ok(one_bits_flat()->contains('1'));
ok(one_bits_flat()->contains('1'));
ok(one_bits_flat()->contains('111'));
ok(! one_bits_flat()->contains('0'));
ok(! one_bits_flat()->contains('10'));
ok(! one_bits_flat()->contains('00'));
ok(! one_bits_flat()->contains('1110'));

{
  my $max_k = 10;
  is_deeply([ FLAT_count_contains(one_bits_flat(),$max_k) ],
            [ (1) x ($max_k+1) ]);
}

#------------------------------------------------------------------------------
# zero_digits_flat(2)

ok(zero_digits_flat(2)->contains(''));
ok(zero_digits_flat(2)->contains('0'));
ok(zero_digits_flat(2)->contains('00'));
ok(zero_digits_flat(2)->contains('000'));
ok(! zero_digits_flat(2)->contains('1'));
ok(! zero_digits_flat(2)->contains('01'));
ok(! zero_digits_flat(2)->contains('11'));
ok(! zero_digits_flat(2)->contains('0001'));

{
  my $max_k = 10;
  is_deeply([ FLAT_count_contains(zero_digits_flat(2),$max_k) ],
            [ (1) x ($max_k+1) ]);
}


#------------------------------------------------------------------------------

exit 0
