#!/usr/bin/perl -w

# Copyright 2016, 2017, 2018, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


use 5.004;
use strict;
use Math::PlanePath::ComplexMinus 117; # v.117 for level_to_n_range()

use FindBin;
use lib $FindBin::Bin;
use MyDragon 'xy_rotate_plus90';
use MyLineSegments;
use MyPlanar ':all';
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $level = 8;
if (@ARGV) {
  $level = shift @ARGV;
}
print "  % generated by tools/$FindBin::Script\n";
print "  \\newcommand{\\MyLevel}{$level}\n";

my $path = Math::PlanePath::ComplexMinus->new;
my ($n_lo, $n_hi) = $path->level_to_n_range($level);


sub points_square {
  my ($x,$y) = @_;
  return [ [2*$x-1, 2*$y-1],
           [2*$x+1, 2*$y-1],
           [2*$x+1, 2*$y+1],
           [2*$x-1, 2*$y+1] ];
}

# all segments of the plain i-1 shape
my $plain_segments  = MyLineSegments->new (name => 'plain');

# all segments of the i-1 shape rotated +90 degrees, and which will be
# reduced to just those not in the plain segments
my $other_segments = MyLineSegments->new (name => 'other');

# segments around rotatable points of the plain i-1 shape, meaning their +90
# rotation is still in the shape
my $rotatable_segments = MyLineSegments->new (name => 'rotatable');

# unit squares of the rotatable points of the plain i-1 shape
my $rotatable_gpc = MyPlanar::points_to_gpc(points_square(0,0));

my $MinusRotA = 0;
foreach my $n ($n_lo .. $n_hi) {
  my ($x,$y) = $path->n_to_xy($n);
  my $points = points_square($x,$y);
  $plain_segments->add_cycle(map {@$_} @$points);

  my ($rx,$ry) = xy_rotate_plus90($x,$y);
  my $rn = $path->xy_to_n($rx,$ry);
  if ($rn <= $n_hi) {
    ### rotatable: "n=$n $x,$y to rn=$rn $rx,$ry add " . join(',',map {@$_} @$points)
    $rotatable_segments->add_cycle(map {@$_} @$points);
    my $gpc = MyPlanar::points_to_gpc($points);
    $rotatable_gpc = Math::Geometry::Planar::GpcClip('UNION',$rotatable_gpc,$gpc);
    $MinusRotA++;
  }

  $points = points_square($rx,$ry);
  ### rotate segments add: "$rx,$ry points " . join(',',map {@$_} @$points)
  $other_segments->add_cycle(map {@$_} @$points);
}

### @rotatables
### plain  length: $plain_segments->length
### rotate length: $other_segments->length
### rotatable   length: $rotatable_segments->length

# remove $rotatable_segments, they will be a filled area
$rotatable_segments->is_subset_of($plain_segments) or die;
$plain_segments->remove_segments($rotatable_segments);
$other_segments->remove_segments($rotatable_segments);

# remove $plain_segments, so $other_segments is rest of rotated copy
$other_segments->remove_segments($plain_segments);
$plain_segments->is_disjoint($other_segments) or die;

my $MinusRotB = 0;
{
  my @contours = Math::Geometry::Planar::Gpc2Polygons($rotatable_gpc);
  my $total_holes = 0;
  foreach my $contour (@contours) {  # Math::Geometry::Planar contour
    ### $contour
    # Second and subsequent polygon of the contour are holes.
    # Think their points already cycle the opposite way as required by pgf
    # default \fill[nonzero rule], but using \fill[even odd rule] works either
    # way.
    $Printwrap = 2;
    printwrap(" \\fill[my rotatable,even odd rule]");
    $Printwrap += 2;
    my $polygons = $contour->polygons;  # arrayref of points arrayrefs
    $total_holes += scalar(@$polygons)-1;
    my $separator = '';
    foreach my $points (@$polygons) {
      ### $points
      printwrap($separator);
      $separator = "\n";
      printwrap_poly($points);
    }
    printwrap(";\n");
    $MinusRotB += $contour->perimeter / 2;
  }
  $total_holes == 0 or die;
}

$Printwrap = 2;
printwrap("\\draw[my plain line] ");
$Printwrap += 2;
$plain_segments->printwrap;
printwrap(";\n");

$Printwrap = 2;
printwrap("\\draw[my other line] ");
$Printwrap += 2;
$other_segments->printwrap;
printwrap(";\n");

print "% GP-Test  MinusRotA($level) == $MinusRotA\n";
print "% GP-Test-Last  MinusRotB_by_pred($level) == $MinusRotB\n";
print "  \\newcommand{\\MyMinusRotA}{$MinusRotA}\n";
print "  \\newcommand{\\MyMinusRotB}{$MinusRotB}\n";
print "  % end generated\n";
exit 0;





__END__

# use 5.004;
# use strict;
# use Math::PlanePath::ComplexMinus 117; # v.117 for level_to_n_range()
# 
# use FindBin;
# use lib $FindBin::Bin;
# use MyDragon 'xy_rotate_plus90', 'xy_rotate_minus90';
# use MyLineSegments;
# use MyPrintwrap ':all';
# 
# # uncomment this to run the ### lines
# # use Smart::Comments;
# 
# 
# my $level = 8;
# print "  % generated by tools/$FindBin::Script\n";
# print "  \\newcommand{\\MyLevel}{$level}\n";
# 
# my $path = Math::PlanePath::ComplexMinus->new;
# my ($n_lo, $n_hi) = $path->level_to_n_range($level);
# 
# 
# my $n_segments = MyLineSegments->new (name => 'n');
# my $other_segments = MyLineSegments->new (name => 'rot');
# my $both_segments = MyLineSegments->new (name => 'both');
# 
# sub rotate {
#   my ($x,$y) = @_;
#   return xy_rotate_plus90($x,$y);
# }
# sub unrotate {
#   my ($x,$y) = @_;
#   return xy_rotate_minus90($x,$y);
# }
# 
# sub xy_to_type {
#   my ($x,$y) = @_;
#   my $n   = $path->xy_to_n($x,$y);
#   my ($rx,$ry) = unrotate($x,$y);
#   my $rot = $path->xy_to_n($rx,$ry);
#   ### xy_to_type(): "$x,$y $n   $rx,$y $rot"
#   if ($n <= $n_hi) {
#     return ($rot <= $n_hi ? 'both' : 'n');
#   } else {
#     return ($rot <= $n_hi ? 'rot' : '');
#   }
# }
# 
# sub boundary_to_segs {
#   my ($a,$b) = @_;
#   if ($a eq 'both') {
#     if ($b eq 'both') { return; }                  # hollow
#     else              { return $both_segments; }   # boundary
#   }
#   if ($a eq 'n') {
#     if ($b eq 'both') { return; }
#     return $n_segments;
#   }
#   if ($a eq 'rot') {
#     if ($b eq 'n')    { return; }                  # n--rot as "n"
#     if ($b eq 'both') { return; }
#     return $other_segments;
#   }
#   die "boundary_to_segs() a=$a b=$b";
# }
# 
# my %both;
# 
# foreach my $n ($n_lo .. $n_hi) {
#   ### $n
#   my ($x,$y) = $path->n_to_xy($n);
#   foreach my $rot (0,1) {
#     my $a = xy_to_type($x,$y);
#     if ($a eq 'both') {
#       $both{"$x,$y"} = 1;
#     }
#     ### at: "$x,$y  a=$a (rot=$rot)"
#     foreach my $segments (boundary_to_segs($a, xy_to_type($x,$y-1))) {
#       $segments->add(2*$x-1,2*$y-1, 2*$x+1,2*$y-1);   # S
#     }
#     foreach my $segments (boundary_to_segs($a, xy_to_type($x,$y+1))) {
#       $segments->add(2*$x-1,2*$y+1, 2*$x+1,2*$y+1);   # N
#     }
#     foreach my $segments (boundary_to_segs($a, xy_to_type($x-1,$y))) {
#       $segments->add(2*$x-1,2*$y-1, 2*$x-1,2*$y+1);   # W
#     }
#     foreach my $segments (boundary_to_segs($a, xy_to_type($x+1,$y))) {
#       $segments->add(2*$x+1,2*$y-1, 2*$x+1,2*$y+1);   # E
#     }
#     ($x,$y) = rotate($x,$y);
#   }
# }
# 
# $Printwrap = 2;
# printwrap("\\fill[my intersect] ");
# $Printwrap += 2;
# $both_segments->printwrap;
# printwrap(";\n");
# 
# $Printwrap = 2;
# printwrap("\\draw[my squares line] ");
# $Printwrap += 2;
# $n_segments->printwrap;
# printwrap(";\n");
# 
# $Printwrap = 2;
# printwrap("\\draw[my rotate line] ");
# $Printwrap += 2;
# $other_segments->printwrap;
# printwrap(";\n");
# 
# my $MinusRotA = scalar keys %both;
# print "% GP-Test  MinusRotA($level) == $MinusRotA\n";
# print "  \\newcommand{\\MyMinusRotA}{$MinusRotA}\n";
# print "  % end generated\n";
# exit 0;
