#!/usr/bin/perl -w

# Copyright 2015, 2016, 2017 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


use 5.004;
use strict;
use Tie::IxHash;
use Math::PlanePath::DragonCurve 117; # v.117 for level_to_n_range()

use FindBin;
use lib $FindBin::Bin;
use MyDragon 'chamfer','rect_corners_sort',
  'xy_rotate_plus90','xy_rotate_minus90';
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $level = 6;

print "  % generated by tools/$FindBin::Script\n";
print "  \\newcommand{\\MyLevel}{$level}\n";
my $path = Math::PlanePath::DragonCurve->new;

my ($n_lo, $n_hi) = $path->level_to_n_range($level+1);
my ($x_hi,$y_hi) = $path->n_to_xy($n_hi);
print "  \\coordinate (start) at (0,0);\n";
print "  \\coordinate (end) at ($x_hi,$y_hi);\n";

sub xyxy_is_twindragon_touch {
  my ($x,$y) = @_;
  ### xyxy_is_twindragon_touch(): "$x, $y"
  {
    my $n = $path->xy_to_n($x,$y);
    ### first: $n
    unless ($n && $n < $n_hi) {
      ### bad n: $n_hi
      return 0;
    }
  }
  ($x,$y) = ($x_hi-$x, $y_hi-$y);
  ### reverse: "$x, $y"
  {
    my $n = $path->xy_to_n($x,$y);
    ### second: $n
    unless ($n && $n < $n_hi) {
      ### bad n: $n_hi
      return 0;
    }
  }
  return 1;
}

sub path_n_is_left_boundary {
  my ($path, $n, $n_limit) = @_;
  my ($x1,$y1) = $path->n_to_xy($n);
  my ($x2,$y2) = $path->n_to_xy($n+1);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  ($dx,$dy) = xy_rotate_plus90($dx,$dy);  # to side
  my $x1s = $x1 + $dx;
  my $y1s = $y1 + $dy;
  my $x2s = $x2 + $dx;
  my $y2s = $y2 + $dy;
  ### consider: "$x1,$y1 to $x2,$y2   and $x1s,$y1s to $x2s,$y2s"
  return (! path_xyxy_is_traversed_either($path, $x1,$y1, $x1s,$y1s, $n_limit)
          || ! path_xyxy_is_traversed_either($path, $x2,$y2, $x2s,$y2s, $n_limit)
          || ! path_xyxy_is_traversed_either($path, $x1s,$y1s, $x2s,$y2s, $n_limit));
}
sub path_xyxy_is_traversed_either {
  my ($path, $x1,$y1, $x2,$y2, $n_limit) = @_;
  my $n = $path->xyxy_to_n_either($x1,$y1, $x2,$y2);
  return defined $n && $n < $n_limit;
}


my %squares;
tie %squares, 'Tie::IxHash';

{
  $Printwrap = 2;
  printwrap("\\draw[my dragon line]");
  $Printwrap += 2;
  foreach my $n ($n_lo .. $n_hi) {
    my ($x,$y) = $path->n_to_xy($n);
    my ($dx,$dy) = $path->n_to_dxdy($n);

    if ($n == $n_lo) {
      my ($dx,$dy) = $path->n_to_dxdy($n);
      my $tx = $x . chamfer($dx);
      my $ty = $y . chamfer($dy);
      printwrap("($tx,$ty)");
    } elsif ($n == $n_hi) {
      my ($sdx,$sdy) = $path->n_to_dxdy($n-1);
      my $sx = $x . chamfer(-$sdx);
      my $sy = $y . chamfer(-$sdy);
      printwrap("--($sx,$sy)");
    } else {
      my ($sdx,$sdy) = $path->n_to_dxdy($n-1);
      my ($dx,$dy) = $path->n_to_dxdy($n);
      my $sx = $x . chamfer(-$sdx);
      my $sy = $y . chamfer(-$sdy);
      my $tx = $x . chamfer($dx);
      my $ty = $y . chamfer($dy);
      printwrap("--($sx,$sy)");
      printwrap("--($tx,$ty)");
    }

    if (path_n_is_left_boundary($path, $n, $n_hi)) {
      my ($lx,$ly) = xy_rotate_plus90($dx,$dy);
      my $x2 = $x + $dx + $lx;
      my $y2 = $y + $dy + $ly;
      ($x,$y, $x2,$y2) = rect_corners_sort ($x,$y, $x2,$y2);
      $squares{"$x,$y"} = 1;
    }
  }

  foreach my $n ($n_lo .. $n_hi) {
    my ($x,$y) = $path->n_to_xy($n);
    ($x,$y) = ($x_hi-$x, $y_hi-$y);

    if ($n == $n_lo) {
      my ($tdx,$tdy) = $path->n_to_dxdy($n);
      ($tdx,$tdy) = (-$tdx,-$tdy);
      my $tx = $x . chamfer($tdx);
      my $ty = $y . chamfer($tdy);
      printwrap("--($tx,$ty)");
    } elsif ($n == $n_hi) {
      my ($sdx,$sdy) = $path->n_to_dxdy($n-1);
      ($sdx,$sdy) = (-$sdx,-$sdy);
      my $sx = $x . chamfer(-$sdx);
      my $sy = $y . chamfer(-$sdy);
      printwrap("--($sx,$sy)");
    } else {
      my ($sdx,$sdy) = $path->n_to_dxdy($n-1);
      my ($tdx,$tdy) = $path->n_to_dxdy($n);
      ($sdx,$sdy) = (-$sdx,-$sdy);
      ($tdx,$tdy) = (-$tdx,-$tdy);
      my $sx = $x . chamfer(-$sdx);
      my $sy = $y . chamfer(-$sdy);
      my $tx = $x . chamfer($tdx);
      my $ty = $y . chamfer($tdy);
      printwrap("--($sx,$sy)");
      printwrap("--($tx,$ty)");
    }
  }
  printwrap("--cycle;\n");
}

{
  print "  \\begin{scope}[my square colour,shift={(.5,.5)}]\n";
  $Printwrap = 4;
  printwrap("\\fill foreach \\p in {");
  $Printwrap += 2;
  foreach my $str (keys %squares) {
    printwrap("($str),");
  }
  printwrap("}\n");
  printwrap("{ \\p +(-\\MySquare,-\\MySquare) rectangle +(\\MySquare,\\MySquare) };\n");

  $Printwrap -= 2;
  printwrap("\\draw[my diameter line] ");
  my $join = '';
  $Printwrap += 2;
  foreach my $str (keys %squares) {
    printwrap("$join($str)");
    $join = '--';
  }
  printwrap(";\n");
  print "  \\end{scope}\n";
}

my $num_squares = scalar(keys %squares);
my $diameter = $num_squares - 1;
print "  \\newcommand{\\MyDiameter}{$diameter}\n";
print "% GP-Test  TAdiameter($level) == $diameter\n";

print "  % end generated\n";
exit 0;
