#!/usr/bin/perl -w

# Copyright 2014, 2015, 2016, 2017, 2018, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.004;
use strict;
use Math::PlanePath::DragonCurve 117; # v.117 for level_to_n_range()
use Tie::IxHash;

use FindBin;
use lib $FindBin::Bin;
use MyDragon ':all';
use MyLineSegments;
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $level = 8;

# left right
# both enclosed
my @side_list = ('left','right');
if (@ARGV) { @side_list = @ARGV; }

print "  % generated by tools/$FindBin::Script ",join(' ',@side_list),"\n";
print "  \\newcommand{\\MyLevel}{$level}\n";

my $path = Math::PlanePath::DragonCurve->new;
my ($n_lo, $n_hi) = $path->level_to_n_range($level);

sub path_xyxy_is_left_boundary {
  my ($path, $x1,$y1, $x2,$y2, $n_limit) = @_;
  ### path_xyxy_is_left_boundary(): "$x1,$y1, $x2,$y2, $n_limit"
  if (! path_xyxy_is_traversed_either($path, $x1,$y1, $x2,$y2, $n_limit)) {
    ### not a segment ...
    return 0;
  }
  my $dx = $x2 - $x1;
  my $dy = $y2 - $y1;
  ($dx,$dy) = xy_rotate_plus90($dx,$dy);  # to left side
  my $x1s = $x1 + $dx;
  my $y1s = $y1 + $dy;
  my $x2s = $x2 + $dx;
  my $y2s = $y2 + $dy;
  if (! path_xyxy_is_traversed_either($path, $x1,$y1, $x1s,$y1s, $n_limit)) {
    ### x1 y1 to s untraversed ...
    return 1;
  }
  if (! path_xyxy_is_traversed_either($path, $x2,$y2, $x2s,$y2s, $n_limit)) {
    ### x2 y2 to s ...
    return 1;
  }
  if (! path_xyxy_is_traversed_either($path, $x1s,$y1s, $x2s,$y2s, $n_limit)) {
    ### s to s ...
    return 1;
  }
  ### not left boundary ...
  return 0;
}

sub path_xyxy_is_right_boundary {
  my ($path, $x1,$y1, $x2,$y2, $n_limit) = @_;
  ### path_xyxy_is_right_boundary(): "$x1,$y1, $x2,$y2, $n_limit"
  if (! path_xyxy_is_traversed_either($path, $x1,$y1, $x2,$y2, $n_limit)) {
    return 0;
  }
  my $dx = $x2 - $x1;
  my $dy = $y2 - $y1;
  ($dx,$dy) = xy_rotate_minus90($dx,$dy);  # to right side
  my $x1s = $x1 + $dx;
  my $y1s = $y1 + $dy;
  my $x2s = $x2 + $dx;
  my $y2s = $y2 + $dy;
  return (! path_xyxy_is_traversed_either($path, $x1,$y1, $x1s,$y1s, $n_limit)
          || ! path_xyxy_is_traversed_either($path, $x2,$y2, $x2s,$y2s, $n_limit)
          || ! path_xyxy_is_traversed_either($path, $x1s,$y1s, $x2s,$y2s, $n_limit));
}

sub path_xyxy_is_traversed_either {
  my ($path, $x1,$y1, $x2,$y2, $n_limit) = @_;
  ### xyxy_is_traversed_either(): "$x1,$y1 to $x2,$y2"
  foreach my $n1 ($path->xy_to_n_list($x1,$y1)) {
    next if defined $n_limit && $n1 > $n_limit;
    foreach my $n2 ($path->xy_to_n_list($x2,$y2)) {
      next if defined $n_limit && $n2 > $n_limit;
      if (abs($n1-$n2) == 1) {
        ### yes: "$n1 to $n2"
        return 1;
      }
    }
  }
  return 0;
}

foreach my $side (@side_list) {
  print "  \\begin{scope}[shift=(pos-$side)]\n";

  {
    my $segments = MyLineSegments->new;
    foreach my $n ($n_lo .. $n_hi-1) {
      my ($x,$y) = $path->n_to_xy($n);
      $segments->add($x,$y, $path->n_to_xy($n+1));
    }
    $Printwrap = 4;
    printwrap("\\draw[my grey]");
    $Printwrap += 2;
    foreach my $line ($segments->lines) {
      my ($x1,$y1, $x2,$y2) = @$line;
      printwrap(" ($x1,$y1)--($x2,$y2)");
    }
    printwrap(";\n");
  }

  my %all;
  my %seen;
  tie %all, 'Tie::IxHash';
  tie %seen, 'Tie::IxHash';
  $all{"0,0"} = 1;
  $seen{"0,0"} = 1;

  foreach my $n ($n_lo+1 .. $n_hi) {
    ### $n
    my ($prev_x,$prev_y) = $path->n_to_xy($n-1);
    my ($x,$y) = $path->n_to_xy($n);
    $all{"$x,$y"} = 1;

    if ($side eq 'left') {
      path_xyxy_is_left_boundary($path, $prev_x,$prev_y, $x,$y, $n_hi)
        or next;
    } elsif ($side eq 'right') {
      path_xyxy_is_right_boundary($path, $prev_x,$prev_y, $x,$y, $n_hi)
        or next;
    } elsif ($side eq 'both' || $side eq 'enclosed') {
      next unless
        (path_xyxy_is_left_boundary($path, $prev_x,$prev_y, $x,$y, $n_hi)
         || path_xyxy_is_right_boundary($path, $prev_x,$prev_y,$x,$y, $n_hi));
    } else {
      die;
    }
    $seen{"$x,$y"} = 1;
  }

  if ($side eq 'enclosed') {
    foreach my $p (keys %all) {
      if (exists $seen{$p}) {
        delete $seen{$p};
      } else {
        $seen{$p} = 1;
      }
    }
  }

  $Printwrap = 4;
  printwrap("\\fill foreach \\p in {");
  $Printwrap += 2;
  my $count = 0;
  foreach my $p (keys %seen) {
    printwrap("($p),");
    $count++;
  }
  printwrap("}\n");
  print "    { \\p circle[] };\n";

  my $func = ($side eq 'left' ? 'PL'
              : $side eq 'right' ? 'PR'
              : $side eq 'both' ? 'PB'
              : $side eq 'enclosed' ? 'PE'
              : die);
  my $name = "\u$side";
  print "  \\end{scope}\n";
  print "  \\newcommand{\\My${name}Count}{$count}\n";
  print "% GP-Test  $func($level)==$count\n";
}

print "  % end generated\n";
exit 0;
