#!/usr/bin/perl -w

# Copyright 2015, 2016, 2017, 2019, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.004;
use strict;
use Math::PlanePath::DragonCurve 117; # v.117 for level_to_n_range()

use FindBin;
use lib $FindBin::Bin;
use MyDragon 'BlobPathLen','Diameter';
use MyLineSegments;
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $level = 9;
if (@ARGV) { $level = shift @ARGV; }

print "  % generated by tools/$FindBin::Script $level\n";
my $path = Math::PlanePath::DragonCurve->new;

my ($n_lo, $n_hi) = $path->level_to_n_range($level);
my ($x_hi,$y_hi) = $path->n_to_xy($n_hi);

$Printwrap = 2;
printwrap("\\node (start) at (0,0) [my dot];\n");
printwrap("\\node (end) at ($x_hi,$y_hi) [my dot];\n");

my $dragon_segments = MyLineSegments->new;
my $path_segments = MyLineSegments->new;
foreach my $n ($n_lo .. $n_hi-1) {
  $dragon_segments->add($path->n_to_xy($n),
                 $path->n_to_xy($n+1));
}

my $Diameter = 0;
{
  # shortest path

  my $north = [0,1];
  my $south = [0,-1];
  my $west = [-1,0];
  my $east = [1,0];

  my @start = ($east, $north, $west, $north, $west, $south, $west);
  my @end = ($west, $south, $west, $north);

  my @bridge = ([ $north, $east, $north], # 8
                [ $north, $west, $north], # 9
                [ $west, $north, $west],  # 10
                [ $west, $south, $west],  # 3
                [ $south, $west, $south], # 4
                [ $south, $east, $south], # 5
                [ $east, $south, $east],  # 6
                [ $east, $north, $east],  # 7
               );


  my @blob = ([ $north, $east ],   # 8    NE
              [ $north ],          # 9    N
              [ $north, $west ],   # 10   NW
              [ $west ],           # 11   W
              #
              [ $south, $west ],   # 12   SW
              [ $south ],          # 5    S
              [ $east, $south ],   # 6    SE
              [ $east ],           # 7    E
             );

  my $x = 0;
  my $y = 0;

  my $move = sub {
    my ($d, $nodraw) = @_;
    my ($dx,$dy) = @$d;
    my $x2 = $x + $dx;
    my $y2 = $y + $dy;
    if (! $dragon_segments->is_traversed($x,$y, $x2,$y2)) {
      die "oops $x, $y no segment to $dx, $dy";
    }
    $dragon_segments->remove($x,$y, $x2,$y2);
    $path_segments->add($x,$y, $x2,$y2);
    # unless ($nodraw) {
    # }
    $x = $x2;
    $y = $y2;
    $Diameter += abs($dx) + abs($dy);
  };

  foreach my $d (@start) {
    ### curve start: $d
    $move->($d);
  }
  foreach my $j (4 .. $level) {
    {
      ### start blob: $j
      my $aref = $blob[$j % 8];
      foreach my $i (1 .. BlobPathLen($j)) {
        my $d = $aref->[$i % scalar(@$aref)];
        $move->($d);
      }
    }
    if ($level > 4) {
      ### start bridge: $j
      my $aref = $bridge[$j % 8];
      foreach my $d (@$aref) {
        $move->($d, 1);
      }
    }
  }

  foreach my $j (reverse 4 .. $level-2) {
    {
      ### end blob: $j
      my $aref = $blob[($j+2) % 8];
      ### pathlen: BlobPathLen($j)
      foreach my $i (1 .. BlobPathLen($j)) {
        my $d = $aref->[$i % scalar(@$aref)];
        $move->($d);
      }
    }
    {
      ### end bridge: $j
      my $aref = $bridge[($j+1) % 8];
      foreach my $d (@$aref) {
        $move->($d, $j>4);
      }
    }
  }

  foreach my $d (@end) {
    ### curve end: $d
    $move->($d);
  }

  ($x == $x_hi && $y == $y_hi)
    or die;

  my $Diameter_formula = Diameter($level);
  if ($Diameter != $Diameter_formula) {
    die "Oops, path length $Diameter but formula $Diameter_formula";
  }
}

foreach my $elem ([$dragon_segments,'my dragon'],
                  [$path_segments,'my path']) {
  my ($segments, $colour) = @$elem;
  $Printwrap = 2;
  printwrap("\\draw[$colour] ");
  $Printwrap += 2;
  $segments->printwrap;
  printwrap(";\n");
}

print "  \\newcommand{\\MyLevel}{$level}\n";
print "  \\newcommand{\\MyDiameter}{$Diameter}\n";
print "% GP-Test  Diameter($level) == $Diameter\n";
print "  % end generated\n";
exit 0;
