\\ Copyright 2020 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.


\\ Work in progress.  Will change wildly.  Mostly working but not
\\ particularly good in places.
\\
\\     bool = interval_contains(t,sub)         contains "sub" interval
\\     bool = interval_contains_constant(t,c)
\\     bool = interval_contains_0(t)
\\     t = interval_neg(t)
\\     t = interval_add(x,y)
\\     t = interval_add_constant(t,c)
\\     t = interval_sub(x,y)
\\     t = interval_mul(x,y)
\\     t = interval_recip(t)
\\
\\     t = interval_union(x,y)
\\     t = interval_union0(t)
\\     t = interval_widen_integer(t)
\\     t = interval_widen_bits(t,k)
\\
\\     vec = interval_fraction_digits(t,base=10,limit='none)
\\     t = interval_fraction_fromdigits(v,point=1,base=10)
\\
\\     t = interval_log(t,num_terms)
\\     t = interval_exp(t,num_terms)
\\     t = interval_solve_bisection(t,func,steps)
\\     
\\
\\ Intervals
\\ ---------
\\
\\ An interval is represented by a 2 element vector
\\
\\     [lo, hi]         interval lo <= value <= hi
\\
\\ All functions accept an interval either way around [lo,hi] or [hi,lo].
\\ This is designed to avoid silently giving wrong results if input data
\\ happens to be the wrong way around (which is all too easy for human
\\ input, especially with fractions or negatives).
\\
\\ The purpose of interval arithmetic is to make calculations with an upper
\\ and lower bound around some actual value which is difficult to calculate
\\ or perhaps unknown.  Many quantities such as sum of a series or limit of
\\ a sequence can be hopelessly impractical to represent in an exact way,
\\ only an upper and lower bound to some target width (accuracy).
\\
\\ lo and hi should generally be integers or rationals.  Floating point
\\ works, but interval endpoints are then subject to floating-point rounding
\\ which defeats the purpose of certainty of upper and lower bounds.
\\
\\ Accuracy
\\ --------
\\
\\ Width = hi - lo of an interval is the accuracy with which the interval
\\ represents some actual value.
\\
\\ Addition and subtraction both add widths.  For example
\\
\\     [100,110] + [5,7]  =  [105, 117]     width 10 + 2 = 12
\\     [100,110] - [5,7]  =  [ 93, 105]     width 10 + 2 = 12
\\
\\ Maybe some idea of width proportional to the value could be more
\\ interesting, so say a range lo to lo*1.0001.
\\
\\ Widening
\\ --------
\\
\\ Sometimes during a calculation, denominators from rationals used can
\\ become large, much larger than relevant for the calculation.  It may be
\\ worth widening intervals to some sort of common denominator.
\\ interval_widen_bits() does this to a specified bit precision (a
\\ denominator 2^k).  This is efficient, but some other common base might be
\\ preferred, such as when targeting decimal digits.
\\
\\ An alternative to widening to bit precision is to keep all quantities in
\\ integers with an implicit scale factor.  Maybe that would be a much
\\ better overall concept.  There's some C libraries for that sort of thing
\\ too isn't there?  Keeping a scale separate from a value requires care to
\\ apply the right scale at the right place.


\\-----------------------------------------------------------------------------
interval_validate(t) =
{
  t[1]<=t[2] || error("interval out of order");
}

interval_contains_0(t) = sign(t[1])*sign(t[2]) <= 0;
{
addhelp(interval_contains_0,
"bool = interval_contains_0(t)
Return 1 if interval t contains 0, or return 0 if not.
Interval endpoints are inclusive, so this is either of t[1],t[2] = 0,
or different signs.");
}

interval_contains_constant(t,c) = vecmin(t) <= c && c <= vecmax(t);
{
addhelp(interval_contains_constant,
"bool = interval_contains_constant(t,c)
Return 1 if interval t contains constant c, or return 0 if not.
Interval endpoints are inclusive.");
}

interval_contains(t,sub) = \
  vecmin(t) <= vecmin(sub) && vecmax(sub) <= vecmax(t);
{
addhelp(interval_contains,
"bool = interval_contains(t,sub)
Return 1 if interval t contains within it all of interval \"sub\", as a
sub-interval, or return 0 if not.");
}

interval_union(x,y) = x=concat(x,y); [vecmin(x), vecmax(x)];
{
addhelp(interval_union,
"t = interval_union(x,y)
x and y are intervals.  Return their union, meaning the smallest and largest
extents of x,y.");
}

interval_union0(t) = t=concat(0,t); [vecmin(t), vecmax(t)];
{
addhelp(interval_union0,
"t = interval_union0(t)
t is an interval.  Return it unioned with 0, so that the bigger or smaller
raised or lowered to 0 if 0 is not already contained.");
}

interval_neg(t) = [ -t[2], -t[1] ];
{
addhelp(interval_neg,
"t = interval_neg(t)
t is an interval.  Return negation -t.");
}

interval_add(x,y) = [vecmin(x)+vecmin(y), vecmax(x)+vecmax(y)];
{
addhelp(interval_add,
"x = interval_add(x,y)
x and y are intervals.  Return sum x + y.

If x and y are both in the normal order [lo,hi] then plain vector addition
\"x+y\" is enough.  But not if opposite [lo,hi] and [hi,lo].  interval_add()
is correct in all cases.");
}

interval_add_constant(x,c) = [x[1]+c, x[2]+c];
{
addhelp(interval_add_constant,
"t = interval_add_constant(t,c)
t is an interval and c is a constant.  Return interval sum t + c.");
}

interval_sub(x,y) = [vecmin(x)-vecmax(y), vecmax(x)-vecmin(y)];
{
addhelp(interval_sub,
"x = interval_sub(x,y)
x and y are intervals.  Return subtraction x - y.");
}

\\ plot(x=-2,2,1/x)
interval_recip(t) = \
  if(interval_contains_0(t), [-1/0,1/0], [ 1/t[2], 1/t[1] ]);
{
addhelp(interval_recip,
"t = interval_recip(t)
t is an interval.  Return the reciprocal interval 1/t.

If t contains 0 (either an endpoint or within) then a division-by-0 error
results.  An infinite interval [-oo,oo] would be possible, but unlikely to
be useful since that is effectively no information at all about the value.");
}

interval_mul(x,y) = [ vecmin(x)*vecmin(y), vecmax(x)*vecmax(y) ];
{
addhelp(interval_mul,
"t = interval_mul(x,y)
x and y are intervals.  Return product x*y.");
}

interval_div(x,y) = interval_mul(x,interval_recip(y));
{
addhelp(interval_div,
"t = interval_div(x,y)
x and y are intervals.  Return division x/y.");
}

interval_pow(t,n) =
{
  if(n%2==0 && interval_contains_0(t),
     [0, vecmax(abs(t))^n],
     vecsort([t[1]^n, t[2]^n]));
}
{
addhelp(interval_pow,
"t = interval_pow(t,n)
t is an interval.  Return power t^n.");
}


\\-----------------------------------------------------------------------------
\\ Widening

interval_widen_integer(t) = [floor(vecmin(t)), ceil(vecmax(t))];
{
addhelp(interval_widen_integer,
"t = interval_widen_integer(t)
t is an interval.  Return it widened it to the next bigger and smaller
integers.  See also interval_widen_bits().");
}

interval_widen_bits(t,k) = my(p=2^k); interval_widen_integer(t*p) / p;
{
addhelp(interval_widen_bits,
"t = interval_widen_bits(t,k)
t is an interval.  Return it widened it to the next bigger and smaller
fraction of the form n/2^k.

This is useful for avoiding large denominators when various terms are added
or multiplied.  Usually k will be some bit position a little beyond the
target precision for a final result.

k=0 is the same as interval_widen_integer().");
}


\\-----------------------------------------------------------------------------
\\ Digits

interval_fraction_digits(t,base=10,limit='none) =
{
  my(l=List([]), f);
  while((limit=='none || #l<limit) && (f=floor(t))[1]==f[2],
        listput(l,f[1]);
        t=(t-f)*base);
  Vec(l);
}
{
addhelp(interval_fraction_digits,
"vec = interval_fraction_digits(t,base=10,limit='none)
t is an interval.  Return a vector of its integer part then fractional
digits, up to the accuracy of t.

vec[1] returned is the integer part of t (floor), then vec[2..] are decimal
digits.  Optional parameter \"base\" gives digits in another base.

Optional parameter \"limit\" is a maximum length for the return vec, so stop
at vec length \"limit\" or the accuracy of t, whichever comes first.
Default limit 'none is no limit (so all of the accuracy of t).");
}

interval_fraction_fromdigits(v,point=1,base=10) =
{
  my(p=1/base^(#v-point));
  v=fromdigits(v,base)*p;
  [v, v+p];
}
{
addhelp(interval_fraction_fromdigits,
"t = interval_fraction_fromdigits(v,point=1,base=10)
v is a vector of digits, with the radix point immediately after v[point],
which are known to be good, but with further digits after v unknown other
than being in the range 0..base-1.  Return an interval for these digits.

The default point=1 and base=10 suits the corresponding defaults from
interval_fraction_digits().  vec[1] in that case is an integer part (so
possibly bigger than \"9\").  Any digits can be outside the normal 0..9, so
long as the unknown part means the interval is v to v + 1/base^point.");
}


\\-----------------------------------------------------------------------------
\\ Continued Fractions

interval_contfrac(t,limit='none) =
{
  my(l=List([]), f);
  while((limit=='none || #l<limit) && (f=floor(t))[1]==f[2],
        listput(l,f[1]);
        t=interval_recip(t-f));
  Vec(l);
}
{
addhelp(interval_contfrac,
"vec = interval_contfrac(t,limit='none)
Return a vector of terms of the continued fraction expansion of interval t,
for those terms known with certainty from the interval.

Optional parameter \"limit\" can be an integer limit for the length of the
return.  So stop at vec length \"limit\", or accuracy of t, whichever comes
first.  Default symbol 'none is no limit (so all of the accuracy of t).");
}


\\-----------------------------------------------------------------------------
\\ Polynomials

interval_subst(poly,var,t) =
{
  \\ print("p "valuation(p,var)" "poldegree(p,var)" "polcoeff(p,1,var));
  sum(i=valuation(p,var),poldegree(p,var),
      vecsort(polcoeff(p,i,var) * interval_pow(value,i)));
}
{
addhelp(interval_subst,
"t = interval_subst(poly,var,t)
poly is a polynomial in variable \"var\".  t is an interval.
Return the interval which is poly evaluated in the range of t.");
}


\\-----------------------------------------------------------------------------
\\ Sums of Series

interval_sumalt(func,num_terms,n_start=0) =
{
  num_terms>=2 || error("interval_sumalt() is for num_terms>=2");
  my(t=[0,0],prev,total=0);
  for(n=n_start,n_start+num_terms-1,
     prev=total;
     total+=func(n));
  vecsort([prev,total]);
}
{
addhelp(interval_sumalt,
"t = interval_sumalt(func,num_terms,n_start)
Return an interval for sum of func(n) for \"num_terms\" of n starting at
n = n_start.

Each call is term = func(n) for a term of the sum.  These terms are expected
to be alternating signs and decreasing magnitude so the sum to num_terms and
num_terms-1 are an interval surrounding what would be an infinite sum.");
}


\\-----------------------------------------------------------------------------
\\ log and exp

interval_log(t,num_terms) =
{
  \\ log(x) = -log(1/x) = -log(1-x/x+1/x) = -log(1-(x-1)/x)
  \\ log(7)
  \\ -log(1/7)
  \\ -log(1 - 6/7)
  t=vecsort(t);
  if(t[1]>=1,
     interval_neg
     (interval_log_1plusx
      (interval_neg
       (interval_div(interval_add_constant(t,-1),
                     t)),
       num_terms))
     ,
     interval_log_1plusx(interval_add_constant(t,-1), num_terms));
}
{
addhelp(interval_log,
"t = interval_log(t,num_terms)
t is an interval.  Return an interval for log(t).

The current implementation uses Mercator's series with \"num_terms\" many
terms (calculated on the reciprocal if t > 2).");
}

\\ log(1+x)
\\ omit sum i>=n  (-1)^n * x^n/n
\\
interval_log_1plusx(t,num_terms) =
{
  \\ for x<0, i=1 lower bound want even num terms
  \\          i=2 upper bound want odd num terms
  \\
  t=vecsort(t);
  num_terms = ceil(num_terms);
  (-1 < t[1] && t[2] < 1) || error("interval_log_1plusx() is for -1 < t < 1");
  vector(#t,i,
         my(x=-t[i],
            L=if(x<0,
                 num_terms + bitand(1,bitxor(num_terms,i)),
                 num_terms),
            p=-1,
            e=sum(n=1,L-1,
                  \\ print("n="n" term ",(p*x)/n);
                  (p*=x)/n));
         \\ print("i="i" L="L" x="x"  e="e"  ending p="p);
         if(x<0, if(i==1,p<0||error()); if(i==2,p>0||error()));

         if(i==1&&x>0,
            \\ lower bound on negative omitted part
            \\ print("  omitted ",1/(L*(1-x))," * ",p*x);  p<0 || error();
            e + p*x/(L*(1-x)),
            e));

  \\ interval_validate(x);
  \\ \\ x[1]>-1 || error("interval_log_1plusx() must have x>=1");
  \\ my(ret=[0,0]);
  \\ for(i=1,num_terms,
  \\    interval_validate(ret);
  \\    my(t=interval_pow(x,i)/i);
  \\    if(i%2==0, t=interval_neg(t));
  \\    interval_validate(t);
  \\    ret += t);
  \\ ret;
}
{
addhelp(interval_log_1plusx,
"t = interval_log_1plusx(t,num_terms)
Return an interval for log(1+t) of interval t.
This is Mercator's series of num_terms, so down to term 1/(n-1)*x^(n-1).
t must be in the range -1 <= t <= 1 which is the series radius of
convergence, and note that convergence is slow for abs(t) near 1.

For positive values in t, the series alternates and upper and lower bounds
are by stopping at an odd or even term (positive or negative).  This uses 1
term additional to num_terms if necessary.

For negative values in t, a simple negative lower bound is made by a
geometric series for the omitted part.");
}

\\ exp('t)
\\ omit n>=L       sum x^n/n!
\\      <=  x^L/L! sum (x/L)^j
\\ x/L < 1
\\ L > x
\\ suminf(j=0, (2/7)^j)
\\ 1/(1 - 2/7)  *1.0
\\ 1/(1 - x/L)
\\ GP-Test  1/(1 - 'x/'L) == 'L/('L - 'x)
\\
interval_exp(t,num_terms) =
{
  \\ for x<0, i=1 lower bound want even num terms
  \\          i=2 upper bound want odd num terms
  \\
  t=vecsort(t);
  num_terms = ceil(num_terms);
  vector(#t,i,
         my(x=t[i],
            L=if(x<0,
                 num_terms + bitnegimply(1,bitxor(num_terms,i)),
                 max(num_terms, floor(x)+1)),
            p=1,
            e=(L>0) + sum(n=1,L-1, p=p*x/n));
         \\ print("i="i" L="L"  e="e"  ending p="p);
         if(x<0, if(i==1,p<0||error()); if(i==2,p>0||error()));

         if(i==2&&x>0,
            \\ upper bound on omitted part
            \\ print("  omitted ",L/(L - x)," * ",p*x);
            e + p*x/(L - x),
            e));
}
{
addhelp(interval_exp,
"t = interval_exp(t,num_terms)
t is an interval.  Return an interval for exp(t).
Currently this uses the Taylor series expansion, of \"num_terms\" many terms.

For negatives in t, the series alternates and an upper or lower bound
endpoint is found by stopping at an even or odd term as appropriate.  This
may be 1 term additional to num_terms if necessary.

For positives in t, the series is increasing and a simple upper bound is
made by a geometric series for the omitted part.  For hi the maximum in t,
this requires continuing until term n with hi/n < 1, so num_terms is
increased if necessary to at least hi many.");
}

\\ term x^n/n!
interval_exp_estimate_num_terms(x,bitpos) =
{
  x=vecmax(x);
  floor(log(x));
}

\\-----------------------------------------------------------------------------

interval_solve_bisection(t,func,steps) =
{
  t=vecsort(t);
  my(ylo=func(t[1]));
  if(ylo==0, t[2]=t[1],
     my(yhi=func(t[2]));
     if(yhi==0, t[1]=t[2],
        if(sign(ylo)==sign(yhi),
           error("interval_solve_bisection() func does not change sign"));
        for(i=1,steps,
           my(tmid=vecsum(t)/2, ymid=func(tmid));
           if(ymid==0, t[1]=t[2]=tmid;break,
              sign(ymid)==sign(ylo), t[1]=tmid,
                                     t[2]=tmid))));
  t;
}
{
addhelp(interval_solve_bisection,
"t = interval_solve_bisection(t,func,steps)
t is in an interval over which func changes sign,
so sign(func(lo)) != sign(func(hi)).
Narrow the interval by taking the midpoint and putting it to either lo or hi
according as sign(func(mid)).  Do this \"steps\" many times and return the
result.

If some func(xmid)==0 is found, then return that as an exact solution found
[xmid,xmid].

This bisection suits func(x) which is continuous and gives a rational result
for a rational x.  The width of interval t will narrow to width/2^steps.

If func(x)==0 at multiple places within t then it's unwise to rely on which
might be narrowed to.  Usually t should begin accurate enough to surround a
single solution, though that's not required.  For polynomial roots, see
polsturm() to find how many solutions in a given range.");
}

\\-----------------------------------------------------------------------------
